# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from botocore.session import get_session


def test_start_medical_scribe_listening_session_removal():
    """start_medical_scribe_listening_session operation removed due to h2 requirement"""
    session = get_session()
    connecthealth = session.create_client('connecthealth', 'us-west-2')
    with pytest.raises(AttributeError):
        connecthealth.start_medical_scribe_listening_session



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

from tests import BaseSessionTest, ClientHTTPStubber, mock


class TestLex(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = 'us-west-2'
        self.client = self.session.create_client('lex-runtime', self.region)
        self.http_stubber = ClientHTTPStubber(self.client)

    def test_unsigned_payload(self):
        params = {
            'botName': 'foo',
            'botAlias': 'bar',
            'userId': 'baz',
            'contentType': 'application/octet-stream',
            'inputStream': b'',
        }

        timestamp = datetime.datetime(
            2017, 3, 22, 0, 0, tzinfo=datetime.timezone.utc
        )

        with mock.patch('botocore.auth.datetime.datetime') as _datetime:
            _datetime.now.return_value = timestamp
            self.http_stubber.add_response(body=b'{}')
            with self.http_stubber:
                self.client.post_content(**params)
                request = self.http_stubber.requests[0]

        # The payload gets added to the string to sign, and then part of the
        # signature. The signature will be part of the authorization header.
        # Since we don't have direct access to the payload signature,
        # we compare the authorization instead.
        authorization = request.headers.get('authorization')

        expected_authorization = (
            b'AWS4-HMAC-SHA256 '
            b'Credential=access_key/20170322/us-west-2/lex/aws4_request, '
            b'SignedHeaders=content-type;host;x-amz-content-sha256;x-amz-date,'
            b' Signature='
            b'7f93fde5c36163dce6ee116fcfebab13474ab903782fea04c00bb1dedc3fc4cc'
        )
        self.assertEqual(authorization, expected_authorization)

        content_header = request.headers.get('x-amz-content-sha256')
        self.assertEqual(content_header, b'UNSIGNED-PAYLOAD')


class TestLexV2(BaseSessionTest):
    def test_start_conversation(self):
        """StartConversation operation removed due to h2 requirement"""
        lexv2 = self.session.create_client('lexv2-runtime', 'us-west-2')
        try:
            lexv2.start_conversation
        except AttributeError:
            pass
        else:
            self.fail(
                'start_conversation shouldn\'t be available on the '
                'lexv2-runtime client.'
            )



# Copyright 2012-2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
from functools import lru_cache
from pathlib import Path

import pytest

from botocore import xform_name
from botocore.compat import HAS_CRT
from botocore.config import Config
from botocore.endpoint_provider import (
    S3_UNREFERENCED_PARAMS,
    EndpointProvider,
)
from botocore.exceptions import (
    BotoCoreError,
    ClientError,
    EndpointResolutionError,
)
from botocore.loaders import Loader
from botocore.parsers import ResponseParserError
from tests import ClientHTTPStubber

ENDPOINT_TESTDATA_DIR = Path(__file__).parent / 'endpoint-rules'
LOADER = Loader()

# For the purpose of the tests in this file, only services for which an
# endpoint ruleset file exists matter. The existence of required endpoint
# ruleset files is asserted for in tests/functional/test_model_completeness.py
ALL_SERVICES = [
    service_name
    for service_name in LOADER.list_available_services(
        type_name='endpoint-rule-set-1'
    )
]


@pytest.fixture(scope='module')
def partitions():
    return LOADER.load_data('partitions')


@lru_cache
def get_endpoint_tests_for_service(service_name):
    file_path = ENDPOINT_TESTDATA_DIR / service_name / 'endpoint-tests-1.json'
    if not file_path.is_file():
        raise FileNotFoundError(
            f'Cannot find endpoint tests file for "{service_name}" at '
            f'path {file_path}'
        )
    with file_path.open('r') as f:
        return json.load(f)


@pytest.mark.parametrize("service_name", ALL_SERVICES)
def test_all_endpoint_tests_exist(service_name):
    """Tests the existence of endpoint-tests-1.json for each service that has
    a ruleset and verifies that content is present."""
    data = get_endpoint_tests_for_service(service_name)
    assert len(data['testCases']) > 0


def assert_all_signing_region_sets_have_length_one(rule):
    """Helper function for test_all_signing_region_sets_have_length_one()"""
    if 'endpoint' in rule:
        authSchemes = (
            rule['endpoint'].get('properties', {}).get('authSchemes', [])
        )
        for authScheme in authSchemes:
            if 'signingRegionSet' in authScheme:
                assert len(authScheme['signingRegionSet']) == 1
    for sub_rule in rule.get('rules', []):
        assert_all_signing_region_sets_have_length_one(sub_rule)


@pytest.mark.parametrize("service_name", ALL_SERVICES)
def test_all_signing_region_sets_have_length_one(service_name):
    """Checks all endpoint rulesets for endpoints that contain an authSchemes
    property with a `signingRegionSet` and asserts that it is a list of
    length 1.

    In theory, `signingRegionSet` could have >1 entries. As of writing this
    test, no service uses >1 entry, the meaning of >1 entry is poorly defined,
    and botocore cannot handle >1 entry. This test exists specifically to
    fail if a ruleset ever uses >1 entry.

    The test also fails for empty lists. While botocore would handle these
    gracefully, the expected behavior for empty `signingRegionSet` lists is
    not defined.
    """
    ruleset = LOADER.load_service_model(service_name, 'endpoint-rule-set-1')
    assert_all_signing_region_sets_have_length_one(ruleset)


def test_assert_all_signing_region_sets_have_length_one():
    """Negative test for to confirm that
    assert_all_signing_region_sets_have_length_one() actually fails when two
    sigingRegionSet entries are present."""
    with pytest.raises(AssertionError):
        assert_all_signing_region_sets_have_length_one(
            {
                "version": "1.0",
                "parameters": {},
                "rules": [
                    {
                        "conditions": [],
                        "endpoint": {
                            "url": "https://foo",
                            "properties": {
                                "authSchemes": [
                                    {
                                        "name": "sigv4a",
                                        "disableDoubleEncoding": True,
                                        "signingRegionSet": ["*", "abc"],
                                        "signingName": "myservice",
                                    }
                                ]
                            },
                            "headers": {},
                        },
                        "type": "endpoint",
                    }
                ],
            }
        )


def iter_all_test_cases():
    for service_name in ALL_SERVICES:
        test_data = get_endpoint_tests_for_service(service_name)
        for test_case in test_data['testCases']:
            yield service_name, test_case


def iter_provider_test_cases_that_produce(endpoints=False, errors=False):
    for service_name, test in iter_all_test_cases():
        input_params = test.get('params', {})
        expected_object = test['expect']
        if endpoints and 'endpoint' in expected_object:
            yield service_name, input_params, expected_object['endpoint']
        if errors and 'error' in expected_object:
            yield service_name, input_params, expected_object['error']


def iter_e2e_test_cases_that_produce(endpoints=False, errors=False):
    for service_name, test in iter_all_test_cases():
        # Not all test cases contain operation inputs for end-to-end tests.
        if 'operationInputs' not in test:
            continue
        # Each test case can contain a list of input sets for the same
        # expected result.
        for op_inputs in test['operationInputs']:
            op_params = op_inputs.get('operationParams', {})
            # Test cases that use invalid bucket names as inputs fail in
            # botocore because botocore validated bucket names before running
            # endpoint resolution.
            if op_params.get('Bucket') in ['bucket name', 'example.com#']:
                continue
            op_name = op_inputs['operationName']
            builtins = op_inputs.get('builtInParams', {})

            expected_object = test['expect']
            if endpoints and 'endpoint' in expected_object:
                expected_endpoint = expected_object['endpoint']
                expected_props = expected_endpoint.get('properties', {})
                expected_authschemes = [
                    auth_scheme['name']
                    for auth_scheme in expected_props.get('authSchemes', [])
                ]
                yield pytest.param(
                    service_name,
                    op_name,
                    op_params,
                    builtins,
                    expected_endpoint,
                    marks=pytest.mark.skipif(
                        'sigv4a' in expected_authschemes and not HAS_CRT,
                        reason="Test case expects sigv4a which requires CRT",
                    ),
                )
            if errors and 'error' in expected_object:
                yield pytest.param(
                    service_name,
                    op_name,
                    op_params,
                    builtins,
                    expected_object['error'],
                )


@pytest.mark.parametrize(
    'service_name, input_params, expected_endpoint',
    iter_provider_test_cases_that_produce(endpoints=True),
)
def test_endpoint_provider_test_cases_yielding_endpoints(
    partitions, service_name, input_params, expected_endpoint
):
    ruleset = LOADER.load_service_model(service_name, 'endpoint-rule-set-1')
    endpoint_provider = EndpointProvider(ruleset, partitions)
    endpoint = endpoint_provider.resolve_endpoint(**input_params)
    assert endpoint.url == expected_endpoint['url']
    assert endpoint.properties == expected_endpoint.get('properties', {})
    assert endpoint.headers == expected_endpoint.get('headers', {})


@pytest.mark.parametrize(
    'service_name, input_params, expected_error',
    iter_provider_test_cases_that_produce(errors=True),
)
def test_endpoint_provider_test_cases_yielding_errors(
    partitions, service_name, input_params, expected_error
):
    ruleset = LOADER.load_service_model(service_name, 'endpoint-rule-set-1')
    endpoint_provider = EndpointProvider(ruleset, partitions)
    with pytest.raises(EndpointResolutionError) as exc_info:
        endpoint_provider.resolve_endpoint(**input_params)
    assert str(exc_info.value) == expected_error


@pytest.mark.parametrize(
    'service_name, op_name, op_params, builtin_params, expected_endpoint',
    iter_e2e_test_cases_that_produce(endpoints=True),
)
def test_end_to_end_test_cases_yielding_endpoints(
    patched_session,
    service_name,
    op_name,
    op_params,
    builtin_params,
    expected_endpoint,
):
    def builtin_overwriter_handler(builtins, **kwargs):
        # must edit builtins dict in place but need to erase all existing
        # entries
        for key in list(builtins.keys()):
            del builtins[key]
        for key, val in builtin_params.items():
            builtins[key] = val

    region = builtin_params.get('AWS::Region', 'us-east-1')
    client = patched_session.create_client(
        service_name,
        region_name=region,
        # endpoint ruleset test cases do not account for host prefixes from the
        # operation model
        config=Config(inject_host_prefix=False),
    )
    client.meta.events.register_last(
        'before-endpoint-resolution', builtin_overwriter_handler
    )
    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(status=418)
        op_fn = getattr(client, xform_name(op_name))
        try:
            op_fn(**op_params)
        except (ClientError, ResponseParserError):
            pass
        assert len(http_stubber.requests) > 0
        actual_url = http_stubber.requests[0].url
        assert actual_url.startswith(expected_endpoint['url']), (
            f"{actual_url} does not start with {expected_endpoint['url']}"
        )


@pytest.mark.parametrize(
    'service_name, op_name, op_params, builtin_params, expected_error',
    iter_e2e_test_cases_that_produce(errors=True),
)
def test_end_to_end_test_cases_yielding_errors(
    patched_session,
    service_name,
    op_name,
    op_params,
    builtin_params,
    expected_error,
):
    def builtin_overwriter_handler(builtins, **kwargs):
        # must edit builtins dict in place but need to erase all existing
        # entries
        for key in list(builtins.keys()):
            del builtins[key]
        for key, val in builtin_params.items():
            builtins[key] = val

    region = builtin_params.get('AWS::Region', 'us-east-1')
    client = patched_session.create_client(service_name, region_name=region)
    client.meta.events.register_last(
        'before-endpoint-resolution', builtin_overwriter_handler
    )
    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(status=418)
        op_fn = getattr(client, xform_name(op_name))
        with pytest.raises(BotoCoreError):
            try:
                op_fn(**op_params)
            except (ClientError, ResponseParserError):
                pass
        assert len(http_stubber.requests) == 0


def _collect_refs_from_rules(rules):
    """Recursively collect all {\"ref\": \"...\"} values from rules."""
    refs = set()
    if isinstance(rules, dict):
        if 'ref' in rules:
            refs.add(rules['ref'])
        for value in rules.values():
            refs.update(_collect_refs_from_rules(value))
    elif isinstance(rules, list):
        for item in rules:
            refs.update(_collect_refs_from_rules(item))
    return refs


@pytest.mark.validates_models
def test_s3_unreferenced_endpoint_ruleset_params():
    """Assert that the set of known S3 endpoint ruleset parameters that are
    not used in the rules hasn't changed. These are set in the
    S3_UNREFERENCED_PARAMS constant to prevent them from influencing the cache.
    """
    ruleset = LOADER.load_service_model('s3', 'endpoint-rule-set-1')
    all_params = set(ruleset['parameters'].keys())
    referenced = _collect_refs_from_rules(ruleset['rules'])
    unreferenced = all_params - referenced
    assert unreferenced == S3_UNREFERENCED_PARAMS, (
        "The set of unused S3 endpoints parameters have changed, examine the "
        "implication of this on endpoint caching behavior before updating"
        "S3_UNREFERENCED_PARAMS."
    )



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
class FirstPluginModule:
    def __init__(self):
        self.invocations = 0

    def register_event(self, client):
        client.meta.events.register('before-call', self.increment_invocations)

    def increment_invocations(self, **kwargs):
        self.invocations += 1


plugin_instance = FirstPluginModule()


def initialize_client_plugin(client):
    plugin_instance.register_event(client)
    return plugin_instance



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io

from tests import ClientHTTPStubber


def create_glacier_client(patched_session):
    return patched_session.create_client('glacier', 'us-west-2')


def test_can_list_vaults_without_account_id(patched_session):
    client = create_glacier_client(patched_session)
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response()
        client.list_vaults()
        assert '/-/vaults' in http_stubber.requests[0].url


def test_can_list_vaults_with_account_id(patched_session):
    client = create_glacier_client(patched_session)
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response()
        client.list_vaults(accountId='foo')
        assert '/foo/vaults' in http_stubber.requests[0].url


def test_can_upload_archive(patched_session):
    client = create_glacier_client(patched_session)
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response(status=201)
        client.upload_archive(
            vaultName='test-vault',
            body=io.BytesIO(b'bytes content'),
        )
        headers = http_stubber.requests[0].headers
        assert 'x-amz-content-sha256' in headers
        assert 'x-amz-sha256-tree-hash' in headers


def test_can_upload_archive_from_bytes(patched_session):
    client = create_glacier_client(patched_session)
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response(status=201)
        client.upload_archive(vaultName='test-vault', body=b'bytes content')
        headers = http_stubber.requests[0].headers
        assert 'x-amz-content-sha256' in headers
        assert 'x-amz-sha256-tree-hash' in headers


def test_can_upload_multipart_part(patched_session):
    client = create_glacier_client(patched_session)
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response(status=204)
        client.upload_multipart_part(
            vaultName='test-vault',
            uploadId='upload-id',
            body=io.BytesIO(b'bytes content'),
        )
        headers = http_stubber.requests[0].headers
        assert 'x-amz-content-sha256' in headers
        assert 'x-amz-sha256-tree-hash' in headers


def test_can_upload_multipart_part_from_bytes(patched_session):
    client = create_glacier_client(patched_session)
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response(status=204)
        client.upload_multipart_part(
            vaultName='test-vault',
            uploadId='upload-id',
            body=b'bytes content',
        )
        headers = http_stubber.requests[0].headers
        assert 'x-amz-content-sha256' in headers
        assert 'x-amz-sha256-tree-hash' in headers


def test_glacier_version_header_added(patched_session):
    client = create_glacier_client(patched_session)
    with ClientHTTPStubber(client) as http_stubber:
        http_stubber.add_response()
        client.list_vaults()
        assert 'x-amz-glacier-version' in http_stubber.requests[0].headers



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io
import unittest
from io import BytesIO

import pytest

from botocore.awsrequest import AWSResponse
from botocore.compat import HAS_CRT
from botocore.config import Config
from botocore.exceptions import (
    AwsChunkedWrapperError,
    FlexibleChecksumError,
    MissingDependencyException,
)
from botocore.httpchecksum import (
    AwsChunkedWrapper,
    Crc32Checksum,
    CrtCrc32cChecksum,
    CrtCrc32Checksum,
    CrtCrc64NvmeChecksum,
    CrtXxhash3Checksum,
    CrtXxhash64Checksum,
    CrtXxhash128Checksum,
    Sha1Checksum,
    Sha256Checksum,
    Sha512Checksum,
    StreamingChecksumBody,
    apply_request_checksum,
    handle_checksum_body,
    resolve_request_checksum_algorithm,
    resolve_response_checksum_algorithms,
)
from botocore.model import OperationModel, StringShape, StructureShape
from tests import get_checksum_cls, mock, requires_crt


class TestHttpChecksumHandlers(unittest.TestCase):
    def _make_operation_model(
        self,
        http_checksum=None,
        streaming_output=False,
        streaming_input=False,
        required=False,
    ):
        operation = mock.Mock(spec=OperationModel)
        if http_checksum is None:
            http_checksum = {}
        operation.http_checksum = http_checksum
        operation.http_checksum_required = required
        operation.has_streaming_output = streaming_output
        operation.has_streaming_input = streaming_input
        if http_checksum and "requestAlgorithmMember" in http_checksum:
            shape = mock.Mock(spec=StringShape)
            shape.serialization = {"name": "x-amz-request-algorithm"}
            operation.input_shape = mock.Mock(spec=StructureShape)
            operation.input_shape.members = {
                http_checksum["requestAlgorithmMember"]: shape
            }
        return operation

    def _make_http_response(
        self,
        body,
        headers=None,
        context=None,
        streaming=False,
    ):
        if context is None:
            context = {}

        if headers is None:
            headers = {}

        http_response = mock.Mock(spec=AWSResponse)
        http_response.raw = BytesIO(body)
        http_response.content = body
        http_response.status_code = 200
        http_response.headers = headers
        response_dict = {
            "headers": http_response.headers,
            "status_code": http_response.status_code,
            "context": context,
        }
        if streaming:
            response_dict["body"] = BytesIO(body)
        else:
            response_dict["body"] = body
        return http_response, response_dict

    def _build_request(self, body):
        request = {
            "headers": {},
            "body": body,
            "context": {
                "client_config": Config(
                    request_checksum_calculation="when_supported",
                )
            },
            "url": "https://example.com",
        }
        return request

    def test_request_checksum_algorithm_no_model(self):
        request = self._build_request(b"")
        operation_model = self._make_operation_model()
        params = {}
        resolve_request_checksum_algorithm(request, operation_model, params)
        self.assertNotIn("checksum", request["context"])

    def test_request_checksum_algorithm_model_default(self):
        operation_model = self._make_operation_model(
            http_checksum={"requestAlgorithmMember": "Algorithm"}
        )

        # Param is not present, crc32 checksum will be set by default
        params = {}
        request = self._build_request(b"")
        resolve_request_checksum_algorithm(request, operation_model, params)
        expected_algorithm = {
            "algorithm": "crc32",
            "in": "header",
            "name": "x-amz-checksum-crc32",
        }
        expected_request_algorithm_header = {
            "name": "x-amz-request-algorithm",
            "value": "CRC32",
        }
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        actual_request_algorithm_header = request["context"]["checksum"][
            "request_algorithm_header"
        ]
        self.assertEqual(actual_algorithm, expected_algorithm)
        self.assertEqual(
            actual_request_algorithm_header, expected_request_algorithm_header
        )

        # Param is present, sha256 checksum will be set
        params = {"Algorithm": "sha256"}
        request = self._build_request(b"")
        resolve_request_checksum_algorithm(request, operation_model, params)
        expected_algorithm = {
            "algorithm": "sha256",
            "in": "header",
            "name": "x-amz-checksum-sha256",
        }
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        self.assertEqual(actual_algorithm, expected_algorithm)

        # Param present but header already set, checksum should be skipped
        params = {"Algorithm": "crc32"}
        request = self._build_request(b"")
        request["headers"]["x-amz-checksum-crc32"] = "foo"
        resolve_request_checksum_algorithm(request, operation_model, params)
        self.assertNotIn("checksum", request["context"])

    def test_request_checksum_algorithm_model_default_streaming(self):
        request = self._build_request(b"")
        operation_model = self._make_operation_model(
            http_checksum={"requestAlgorithmMember": "Algorithm"},
            streaming_input=True,
        )

        # Param is not present, crc32 checksum will be set
        params = {}
        resolve_request_checksum_algorithm(request, operation_model, params)
        expected_algorithm = {
            "algorithm": "crc32",
            "in": "trailer",
            "name": "x-amz-checksum-crc32",
        }
        expected_request_algorithm_header = {
            "name": "x-amz-request-algorithm",
            "value": "CRC32",
        }
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        actual_request_algorithm_header = request["context"]["checksum"][
            "request_algorithm_header"
        ]
        self.assertEqual(actual_algorithm, expected_algorithm)
        self.assertEqual(
            actual_request_algorithm_header, expected_request_algorithm_header
        )

        # Param is present, sha256 checksum will be set in the trailer
        params = {"Algorithm": "sha256"}
        resolve_request_checksum_algorithm(request, operation_model, params)
        expected_algorithm = {
            "algorithm": "sha256",
            "in": "trailer",
            "name": "x-amz-checksum-sha256",
        }
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        self.assertEqual(actual_algorithm, expected_algorithm)

        # Trailer should not be used for http endpoints
        request = self._build_request(b"")
        request["url"] = "http://example.com"
        resolve_request_checksum_algorithm(request, operation_model, params)
        expected_algorithm = {
            "algorithm": "sha256",
            "in": "header",
            "name": "x-amz-checksum-sha256",
        }
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        self.assertEqual(actual_algorithm, expected_algorithm)

    def test_request_checksum_algorithm_s3_signature_version_input_streaming(
        self,
    ):
        config = Config(signature_version="s3")
        request = self._build_request(b"")
        request["context"]["client_config"] = config
        operation_model = self._make_operation_model(
            http_checksum={"requestChecksumRequired": True},
            streaming_input=True,
        )

        params = {}
        resolve_request_checksum_algorithm(request, operation_model, params)
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        # Operations with streaming input using the "s3" signature_version should send
        # checksums in the header, not trailer.
        expected_algorithm = {
            "algorithm": "crc32",
            "in": "header",
            "name": "x-amz-checksum-crc32",
        }
        self.assertEqual(actual_algorithm, expected_algorithm)

    def test_request_checksum_algorithm_presigned_request(self):
        request = self._build_request(b"")
        request["context"]["is_presign_request"] = True
        operation_model = self._make_operation_model(
            http_checksum={"requestChecksumRequired": True},
        )

        params = {}
        resolve_request_checksum_algorithm(request, operation_model, params)
        # Presigned requests shouldn't use flexible checksums by default.
        self.assertNotIn("checksum", request["context"])

    def test_request_checksum_algorithm_model_unsupported_algorithm(self):
        request = self._build_request(b"")
        operation_model = self._make_operation_model(
            http_checksum={"requestAlgorithmMember": "Algorithm"},
        )
        params = {"Algorithm": "sha256"}

        with self.assertRaises(FlexibleChecksumError):
            resolve_request_checksum_algorithm(
                request, operation_model, params, supported_algorithms=[]
            )

    def test_request_checksum_algorithm_model_legacy_crc32(self):
        request = self._build_request(b"")
        operation_model = self._make_operation_model(required=True)
        params = {}

        resolve_request_checksum_algorithm(request, operation_model, params)
        expected_algorithm = {
            "algorithm": "crc32",
            "in": "header",
            "name": "x-amz-checksum-crc32",
        }
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        self.assertEqual(actual_algorithm, expected_algorithm)

    def test_request_checksum_algorithm_model_new_crc32(self):
        request = self._build_request(b"")
        operation_model = self._make_operation_model(
            http_checksum={"requestChecksumRequired": True}
        )
        params = {}

        resolve_request_checksum_algorithm(request, operation_model, params)
        actual_algorithm = request["context"]["checksum"]["request_algorithm"]
        expected_algorithm = {
            "algorithm": "crc32",
            "in": "header",
            "name": "x-amz-checksum-crc32",
        }
        self.assertEqual(actual_algorithm, expected_algorithm)

    def test_apply_request_checksum_handles_no_checksum_context(self):
        request = self._build_request(b"")
        apply_request_checksum(request)
        # Build another request and assert the original request is the same
        expected_request = self._build_request(b"")
        self.assertEqual(request["headers"], expected_request["headers"])
        self.assertEqual(request["body"], expected_request["body"])
        self.assertEqual(request["url"], expected_request["url"])

    def test_apply_request_checksum_handles_invalid_context(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "http-trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        with self.assertRaises(FlexibleChecksumError):
            apply_request_checksum(request)

    def test_apply_request_checksum_flex_header_bytes(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "header",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        apply_request_checksum(request)
        self.assertIn("x-amz-checksum-crc32", request["headers"])

    def test_apply_request_checksum_flex_header_readable(self):
        request = self._build_request(BytesIO(b""))
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "header",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        apply_request_checksum(request)
        self.assertIn("x-amz-checksum-crc32", request["headers"])

    def test_apply_request_checksum_flex_header_explicit_digest(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "header",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        request["headers"]["x-amz-checksum-crc32"] = "foo"
        apply_request_checksum(request)
        # The checksum should not have been modified
        self.assertEqual(request["headers"]["x-amz-checksum-crc32"], "foo")

    def test_apply_request_checksum_flex_trailer_bytes(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        apply_request_checksum(request)
        self.assertNotIn("x-amz-checksum-crc32", request["headers"])
        self.assertIsInstance(request["body"], AwsChunkedWrapper)

    def test_apply_request_checksum_flex_trailer_readable(self):
        request = self._build_request(BytesIO(b""))
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        apply_request_checksum(request)
        self.assertNotIn("x-amz-checksum-crc32", request["headers"])
        self.assertIsInstance(request["body"], AwsChunkedWrapper)

    def test_apply_request_checksum_flex_trailer_non_seekable_with_length(
        self,
    ):
        class NonSeekableStream(io.RawIOBase):
            def __init__(self, body):
                self._body = BytesIO(body)

            def read(self, size=-1):
                return self._body.read(size)

            def readable(self):
                return True

        request = self._build_request(NonSeekableStream(b"hello world"))
        request["headers"]["Content-Length"] = "11"
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        apply_request_checksum(request)
        self.assertNotIn("Content-Length", request["headers"])
        self.assertEqual(
            request["headers"]["X-Amz-Decoded-Content-Length"], "11"
        )
        self.assertIsInstance(request["body"], AwsChunkedWrapper)

    def test_apply_request_checksum_flex_header_trailer_explicit_digest(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        request["headers"]["x-amz-checksum-crc32"] = "foo"
        apply_request_checksum(request)
        # The checksum should not have been modified
        self.assertEqual(request["headers"]["x-amz-checksum-crc32"], "foo")
        # The body should not have been wrapped
        self.assertIsInstance(request["body"], bytes)

    def test_apply_request_checksum_content_encoding_preset(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        request["headers"]["Content-Encoding"] = "foo"
        apply_request_checksum(request)
        # The content encoding should only have been appended
        self.assertEqual(
            request["headers"]["Content-Encoding"], "foo,aws-chunked"
        )

    def test_apply_request_checksum_content_encoding_default(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            }
        }
        apply_request_checksum(request)
        self.assertEqual(request["headers"]["Content-Encoding"], "aws-chunked")

    def test_apply_request_checksum_extra_headers(self):
        request = self._build_request(b"")
        request["context"]["checksum"] = {
            "request_algorithm": {
                "in": "trailer",
                "algorithm": "crc32",
                "name": "x-amz-checksum-crc32",
            },
            "request_algorithm_header": {
                "name": "foo",
                "value": "bar",
            },
        }
        apply_request_checksum(request)
        self.assertEqual(request["headers"]["foo"], "bar")

    def test_response_checksum_algorithm_no_model(self):
        request = self._build_request(b"")
        operation_model = self._make_operation_model()
        params = {}
        resolve_response_checksum_algorithms(request, operation_model, params)
        self.assertNotIn("checksum", request["context"])

    def test_response_checksum_algorithm_model_default(self):
        request = self._build_request(b"")
        operation_model = self._make_operation_model(
            http_checksum={
                "responseAlgorithms": [
                    "crc32",
                    "sha1",
                    "sha256",
                    "sha512",
                    "md5",
                    "xxhash64",
                    "xxhash3",
                    "xxhash128",
                ],
                "requestValidationModeMember": "ChecksumMode",
            }
        )

        # Param is not present, no algorithms will be set
        params = {}
        resolve_response_checksum_algorithms(request, operation_model, params)
        self.assertNotIn("checksum", request["context"])

        # Param is present, algorithms will be set
        params = {"ChecksumMode": "enabled"}
        resolve_response_checksum_algorithms(
            request,
            operation_model,
            params,
            supported_algorithms=["sha1", "sha256"],
        )
        # CRC32 should have been filtered it out as it was not supported
        expected_algorithms = ["sha1", "sha256"]
        actual_algorithms = request["context"]["checksum"][
            "response_algorithms"
        ]
        self.assertEqual(actual_algorithms, expected_algorithms)

    def test_handle_checksum_body_checksum(self):
        context = {"checksum": {"response_algorithms": ["sha1", "crc32"]}}
        headers = {"x-amz-checksum-crc32": "DUoRhQ=="}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
        )
        operation_model = self._make_operation_model()
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )
        body = response_dict["body"]
        self.assertEqual(body, b"hello world")
        algorithm = response_dict["context"]["checksum"]["response_algorithm"]
        self.assertEqual(algorithm, "crc32")

        headers = {"x-amz-checksum-crc32": "WrOonG=="}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
        )
        with self.assertRaises(FlexibleChecksumError):
            handle_checksum_body(
                http_response,
                response_dict,
                context,
                operation_model,
            )

        # This header should not be checked, we won't calculate a checksum
        # but a proper body should still come out at the end
        headers = {"x-amz-checksum-foo": "FOO=="}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
        )
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )
        body = response_dict["body"]
        self.assertEqual(body, b"hello world")
        algorithm = response_dict["context"]["checksum"]["response_algorithm"]
        self.assertEqual(algorithm, "crc32")

    def test_handle_checksum_body_checksum_streaming(self):
        context = {"checksum": {"response_algorithms": ["sha1", "crc32"]}}
        headers = {"x-amz-checksum-crc32": "DUoRhQ=="}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
            streaming=True,
        )
        operation_model = self._make_operation_model(streaming_output=True)
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )
        body = response_dict["body"]
        self.assertEqual(body.read(), b"hello world")
        algorithm = response_dict["context"]["checksum"]["response_algorithm"]
        self.assertEqual(algorithm, "crc32")

        headers = {"x-amz-checksum-crc32": "WrOonG=="}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
            streaming=True,
        )
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )
        body = response_dict["body"]
        with self.assertRaises(FlexibleChecksumError):
            body.read()

        # This header should not be checked, we won't calculate a checksum
        # but a proper body should still come out at the end
        headers = {"x-amz-checksum-foo": "FOOO=="}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
            streaming=True,
        )
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )
        body = response_dict["body"]
        self.assertEqual(body.read(), b"hello world")
        algorithm = response_dict["context"]["checksum"]["response_algorithm"]
        self.assertEqual(algorithm, "crc32")

    def test_handle_checksum_body_checksum_skip_non_streaming(self):
        context = {"checksum": {"response_algorithms": ["sha1", "crc32"]}}
        # S3 will return checksums over the checksums of parts which are a
        # special case that end with -#. These cannot be validated and are
        # instead skipped
        headers = {"x-amz-checksum-crc32": "FOOO==-123"}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
        )
        operation_model = self._make_operation_model()
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )
        body = response_dict["body"]
        self.assertEqual(body, b"hello world")

    def test_handle_checksum_body_checksum_skip_streaming(self):
        context = {"checksum": {"response_algorithms": ["sha1", "crc32"]}}
        # S3 will return checksums over the checksums of parts which are a
        # special case that end with -#. These cannot be validated and are
        # instead skipped
        headers = {"x-amz-checksum-crc32": "FOOO==-123"}
        http_response, response_dict = self._make_http_response(
            b"hello world",
            headers=headers,
            context=context,
            streaming=True,
        )
        operation_model = self._make_operation_model(streaming_output=True)
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )
        body = response_dict["body"]
        self.assertEqual(body.read(), b"hello world")


@unittest.skipIf(HAS_CRT, "Error only expected when CRT is not available")
@pytest.mark.parametrize(
    "algorithm,expected_error_text",
    [
        ("crc32c", "Using CRC32C requires an additional dependency"),
        ("crc64nvme", "Using CRC64NVME requires an additional dependency"),
        ("xxhash64", "Using XXHASH64 requires an additional dependency"),
        ("xxhash3", "Using XXHASH3 requires an additional dependency"),
        ("xxhash128", "Using XXHASH128 requires an additional dependency"),
    ],
)
def test_request_checksum_algorithm_no_crt_unsupported(
    algorithm, expected_error_text
):
    operation = mock.Mock(spec=OperationModel)
    operation.http_checksum = {"requestAlgorithmMember": "Algorithm"}
    operation.http_checksum_required = False
    operation.has_streaming_input = False
    shape = mock.Mock(spec=StringShape)
    shape.serialization = {"name": "x-amz-request-algorithm"}
    operation.input_shape = mock.Mock(spec=StructureShape)
    operation.input_shape.members = {"Algorithm": shape}
    request = {
        "headers": {},
        "body": b"",
        "context": {
            "client_config": Config(
                request_checksum_calculation="when_supported",
            )
        },
        "url": "https://example.com",
    }
    params = {"Algorithm": algorithm}
    with pytest.raises(MissingDependencyException, match=expected_error_text):
        resolve_request_checksum_algorithm(request, operation, params)


class TestAwsChunkedWrapper(unittest.TestCase):
    def test_single_chunk_body(self):
        # Test a small body that fits in a single chunk
        bytes = BytesIO(b"abcdefghijklmnopqrstuvwxyz")
        wrapper = AwsChunkedWrapper(bytes)
        body = wrapper.read()
        expected = b"1a\r\nabcdefghijklmnopqrstuvwxyz\r\n0\r\n\r\n"
        self.assertEqual(body, expected)

    def test_multi_chunk_body(self):
        # Test a body that requires multiple chunks
        bytes = BytesIO(b"abcdefghijklmnopqrstuvwxyz")
        wrapper = AwsChunkedWrapper(bytes, chunk_size=10)
        body = wrapper.read()
        expected = (
            b"a\r\n"
            b"abcdefghij\r\n"
            b"a\r\n"
            b"klmnopqrst\r\n"
            b"6\r\n"
            b"uvwxyz\r\n"
            b"0\r\n\r\n"
        )  # fmt: skip
        self.assertEqual(body, expected)

    def test_read_returns_less_data(self):
        class OneLessBytesIO(BytesIO):
            def read(self, size=-1):
                # Return 1 less byte than was asked for
                return super().read(size - 1)

        bytes = OneLessBytesIO(b"abcdefghijklmnopqrstuvwxyz")
        wrapper = AwsChunkedWrapper(bytes, chunk_size=10)
        body = wrapper.read()
        # NOTE: This particular body is not important, but it is important that
        # the actual size of the chunk matches the length sent which may not
        # always be the configured chunk_size if the read does not return that
        # much data.
        expected = (
            b"9\r\n"
            b"abcdefghi\r\n"
            b"9\r\n"
            b"jklmnopqr\r\n"
            b"8\r\n"
            b"stuvwxyz\r\n"
            b"0\r\n\r\n"
        )  # fmt: skip
        self.assertEqual(body, expected)

    def test_single_chunk_body_with_checksum(self):
        wrapper = AwsChunkedWrapper(
            BytesIO(b"hello world"),
            checksum_cls=Crc32Checksum,
            checksum_name="checksum",
        )
        body = wrapper.read()
        expected = b"b\r\nhello world\r\n0\r\nchecksum:DUoRhQ==\r\n\r\n"
        self.assertEqual(body, expected)

    def test_multi_chunk_body_with_checksum(self):
        wrapper = AwsChunkedWrapper(
            BytesIO(b"hello world"),
            chunk_size=5,
            checksum_cls=Crc32Checksum,
            checksum_name="checksum",
        )
        body = wrapper.read()
        expected = (
            b"5\r\n"
            b"hello\r\n"
            b"5\r\n"
            b" worl\r\n"
            b"1\r\n"
            b"d\r\n"
            b"0\r\n"
            b"checksum:DUoRhQ==\r\n\r\n"
        )  # fmt: skip
        self.assertEqual(body, expected)

    def test_multi_chunk_body_with_checksum_iter(self):
        wrapper = AwsChunkedWrapper(
            BytesIO(b"hello world"),
            chunk_size=5,
            checksum_cls=Crc32Checksum,
            checksum_name="checksum",
        )
        expected_chunks = [
            b"5\r\nhello\r\n",
            b"5\r\n worl\r\n",
            b"1\r\nd\r\n",
            b"0\r\nchecksum:DUoRhQ==\r\n\r\n",
        ]
        self.assertListEqual(expected_chunks, list(wrapper))

    def test_wrapper_can_be_reset(self):
        wrapper = AwsChunkedWrapper(
            BytesIO(b"hello world"),
            chunk_size=5,
            checksum_cls=Crc32Checksum,
            checksum_name="checksum",
        )
        first_read = wrapper.read()
        self.assertEqual(b"", wrapper.read())
        wrapper.seek(0)
        second_read = wrapper.read()
        self.assertEqual(first_read, second_read)
        self.assertIn(b"checksum:DUoRhQ==", first_read)

    def test_wrapper_can_only_seek_to_start(self):
        wrapper = AwsChunkedWrapper(BytesIO())
        with self.assertRaises(AwsChunkedWrapperError):
            wrapper.seek(1)
        with self.assertRaises(AwsChunkedWrapperError):
            wrapper.seek(0, whence=1)
        with self.assertRaises(AwsChunkedWrapperError):
            wrapper.seek(1, whence=2)


_CHECKSUM_DIGEST_CASES = [
    (Crc32Checksum, "DUoRhQ=="),
    (Sha1Checksum, "Kq5sNclPz7QV2+lfQIuc6R7oRu0="),
    (Sha256Checksum, "uU0nuZNNPgilLlLX2n2r+sSE7+N6U4DukIj3rOLvzek="),
    (
        Sha512Checksum,
        "MJ7MSJwS1utMxA9QyQLytNDtd+5RGnx6m808qG1M2G+YndNbxf9JlnDaNCVbRbDP2DDoH2Bdz33FVC6TrpzXbw==",
    ),
]
if HAS_CRT:
    _CHECKSUM_DIGEST_CASES.extend(
        [
            (CrtCrc32Checksum, "DUoRhQ=="),
            (CrtCrc32cChecksum, "yZRlqg=="),
            (CrtCrc64NvmeChecksum, "jSnVw/bqjr4="),
            (CrtXxhash64Checksum, "RatnNLIeaWg="),
            (CrtXxhash3Checksum, "1Eex6kDmmIs="),
            (CrtXxhash128Checksum, "340J6T+HSQCpm4d1zBW2xw=="),
        ]
    )


@pytest.mark.parametrize(
    "checksum_cls,expected_digest", _CHECKSUM_DIGEST_CASES
)
def test_checksum_digest(checksum_cls, expected_digest):
    checksum = checksum_cls()
    checksum.update(b"hello world")
    assert checksum.b64digest() == expected_digest


@requires_crt()
@pytest.mark.parametrize(
    "algorithm,expected_class",
    [
        ("crc32", CrtCrc32Checksum),
        ("crc32c", CrtCrc32cChecksum),
        ("crc64nvme", CrtCrc64NvmeChecksum),
        ("xxhash64", CrtXxhash64Checksum),
        ("xxhash3", CrtXxhash3Checksum),
        ("xxhash128", CrtXxhash128Checksum),
    ],
)
def test_crt_checksum_available(algorithm, expected_class):
    actual_cls = get_checksum_cls(algorithm)
    assert actual_cls == expected_class


class TestStreamingChecksumBody(unittest.TestCase):
    def setUp(self):
        self.raw_bytes = b"hello world"
        self.fake_body = BytesIO(self.raw_bytes)
        self._make_wrapper("DUoRhQ==")

    def _make_wrapper(self, checksum):
        self.wrapper = StreamingChecksumBody(
            self.fake_body,
            None,
            Crc32Checksum(),
            checksum,
        )

    def test_basic_read_good(self):
        actual = self.wrapper.read()
        self.assertEqual(actual, self.raw_bytes)

    def test_many_reads_good(self):
        actual = b""
        actual += self.wrapper.read(5)
        actual += self.wrapper.read(5)
        actual += self.wrapper.read(1)
        self.assertEqual(actual, self.raw_bytes)

    def test_basic_read_bad(self):
        self._make_wrapper("duorhq==")
        with self.assertRaises(FlexibleChecksumError):
            self.wrapper.read()

    def test_many_reads_bad(self):
        self._make_wrapper("duorhq==")
        self.wrapper.read(5)
        self.wrapper.read(6)
        # Whole body has been read, next read signals the end of the stream and
        # validates the checksum of the body contents read
        with self.assertRaises(FlexibleChecksumError):
            self.wrapper.read(1)

    def test_handles_variable_padding(self):
        # This digest is equivalent but with more padding
        self._make_wrapper("DUoRhQ=====")
        actual = self.wrapper.read()
        self.assertEqual(actual, self.raw_bytes)

    def test_iter_raises_error(self):
        self._make_wrapper("duorhq==")
        with self.assertRaises(FlexibleChecksumError):
            for chunk in self.wrapper:
                pass

    def test_readinto_good(self):
        chunk = bytearray(6)
        self.assertEqual(6, self.wrapper.readinto(chunk))
        self.assertEqual(chunk, bytearray(b"hello "))
        self.assertEqual(5, self.wrapper.readinto(chunk))
        # Note the trailing space here comes from the fact we've only got 5
        # bytes left to read from the stream into a 6 byte buffer, so it leaves
        # the last byte untouched, which is the space character from the
        # previous read.
        self.assertEqual(chunk, bytearray(b"world "))
        # Whole body has been read, next read signals the end of the stream and
        # validates the checksum of the body contents read
        self.wrapper.readinto(chunk)

    def test_readinto_bad(self):
        self._make_wrapper("duorhq==")
        chunk = bytearray(6)
        self.assertEqual(6, self.wrapper.readinto(chunk))
        self.assertEqual(chunk, bytearray(b"hello "))
        self.assertEqual(5, self.wrapper.readinto(chunk))
        self.assertEqual(chunk, bytearray(b"world "))
        # Whole body has been read, next read signals the end of the stream and
        # validates the checksum of the body contents read
        with self.assertRaises(FlexibleChecksumError):
            self.wrapper.readinto(chunk)

    def test_readinto_zero_bytes(self):
        # Test that readinto returns 0 when 0 bytes are requested
        chunk = bytearray(0)
        self.assertEqual(0, self.wrapper.readinto(chunk))
        self.assertEqual(chunk, bytearray(b""))
        # Whole body has been read, next read signals the end of the stream and
        # validates the checksum of the body contents read
        self.wrapper.readinto(chunk)



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore import utils
from botocore.session import get_session
from tests import ClientHTTPStubber


class RecursivePluginModule:
    """A mock recursive plugin for testing nested client creation."""

    def __init__(self):
        self.called = False

    def register_event(self, client):
        client.meta.events.register(
            'before-call.dynamodb.*', self.create_client
        )

    def create_client(self, **kwargs):
        self.called = True
        session = get_session()
        session.set_credentials('key', 'secret')
        client = utils.create_nested_client(
            session, "dynamodb", region_name="us-west-2"
        )
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=200, body=b'')
            client.list_tables()


plugin_instance = RecursivePluginModule()


def initialize_client_plugin(client):
    plugin_instance.register_event(client)
    return plugin_instance



"""Temporary test suite for new updated retry behavior.

These tests validate the updated retry behavior for standard retries,
which is currently gated behind an internal flag and not yet available
to external users. Once the changes are validated internally and released
publicly, these tests will replace the corresponding tests in
test_standard.py and this file will be removed.
"""

from collections import Counter

import pytest

from botocore import configprovider
from botocore.awsrequest import AWSResponse
from botocore.exceptions import ReadTimeoutError
from botocore.retries import quota, standard
from tests import BaseEnvVar, mock, unittest


@mock.patch('botocore.retries.standard.NEW_RETRIES_ENABLED', True)
class TestExponentialBackoff(unittest.TestCase):
    def setUp(self):
        self.random = lambda: 1
        self.backoff = standard.ExponentialBackoff(
            max_backoff=20, random=self.random
        )

    def test_range_of_exponential_backoff(self):
        backoffs = [
            self.backoff.delay_amount(standard.RetryContext(attempt_number=i))
            for i in range(1, 12)
        ]
        # Note that we're capped at 20 which is our max backoff.
        # Cap kicks in at attempt 10
        self.assertEqual(
            backoffs, [0.05, 0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8, 20, 20]
        )

    def test_exponential_backoff_with_jitter(self):
        backoff = standard.ExponentialBackoff()
        backoffs = [
            backoff.delay_amount(standard.RetryContext(attempt_number=3))
            for i in range(10)
        ]
        # For attempt number 3, we should have a max value of 0.2 (0.05 * 2 ^ 2),
        # so we can assert all the backoff values are within that range.
        # 0.05 is the default non-throttling scale
        for x in backoffs:
            self.assertTrue(0 <= x <= 0.2)

    def test_uniform_rand_dist_on_max_attempts(self):
        backoff = standard.ExponentialBackoff()
        num_datapoints = 10_000
        backoffs = [
            backoff.delay_amount(standard.RetryContext(attempt_number=10))
            for i in range(num_datapoints)
        ]
        self._assert_looks_like_uniform_distribution(backoffs)

    def _assert_looks_like_uniform_distribution(self, backoffs):
        histogram = Counter(int(el) for el in backoffs)
        expected_value = len(backoffs) / len(histogram)
        # This is an arbitrarily chosen tolerance, but we're being fairly
        # lenient here and giving a 20% tolerance.  We're only interested
        # in cases where it's obviously broken and not a uniform distribution.
        tolerance = 0.20
        low = expected_value - (expected_value * tolerance)
        high = expected_value + (expected_value * tolerance)
        out_of_range = [
            str(i) for i in histogram.values() if not low <= i <= high
        ]
        if out_of_range:
            raise AssertionError(
                "Backoff values outside of uniform distribution range "
                f"({low} - {high}): {', '.join(out_of_range)}"
            )


@mock.patch('botocore.retries.standard.NEW_RETRIES_ENABLED', True)
class TestRetryQuotaChecker(unittest.TestCase):
    def setUp(self):
        self.quota = quota.RetryQuota(500)
        self.throttling_detector = standard.ThrottlingErrorDetector(
            standard.RetryEventAdapter()
        )
        self.quota_checker = standard.RetryQuotaChecker(
            self.quota, self.throttling_detector
        )
        self.request_context = {}

    def create_context(
        self,
        is_timeout_error=False,
        status_code=200,
        is_throttling_error=False,
    ):
        caught_exception = None
        parsed_response = {}
        if is_timeout_error:
            caught_exception = ReadTimeoutError(endpoint_url='https://foo')
        if is_throttling_error:
            status_code = 400
            parsed_response = {'Error': {'Code': 'Throttling'}}
        http_response = AWSResponse(
            status_code=status_code, raw=None, headers={}, url='https://foo/'
        )
        context = standard.RetryContext(
            attempt_number=1,
            request_context=self.request_context,
            caught_exception=caught_exception,
            http_response=http_response,
            parsed_response=parsed_response,
        )
        return context

    def test_can_acquire_quota_for_throttling_error(self):
        self.assertTrue(
            self.quota_checker.acquire_retry_quota(
                self.create_context(is_throttling_error=True)
            )
        )
        self.assertEqual(self.request_context['retry_quota_capacity'], 5)

    def test_can_acquire_quota_non_timeout_error(self):
        self.assertTrue(
            self.quota_checker.acquire_retry_quota(self.create_context())
        )
        self.assertEqual(self.request_context['retry_quota_capacity'], 14)

    def test_can_acquire_quota_for_timeout_error(self):
        self.assertTrue(
            self.quota_checker.acquire_retry_quota(
                self.create_context(is_timeout_error=True)
            )
        )
        self.assertEqual(self.request_context['retry_quota_capacity'], 14)

    def test_can_release_quota_based_on_context_value_on_success(self):
        context = self.create_context()
        # This is where we had to retry the request but eventually
        # succeeded.
        http_response = self.create_context(status_code=200).http_response
        self.assertTrue(self.quota_checker.acquire_retry_quota(context))
        self.assertEqual(self.quota.available_capacity, 486)
        self.quota_checker.release_retry_quota(
            context.request_context, http_response=http_response
        )
        self.assertEqual(self.quota.available_capacity, 500)

    def test_can_release_quota_when_succeed_after_throttling_error(self):
        context = self.create_context(is_throttling_error=True)
        http_response = self.create_context(status_code=200).http_response
        self.assertTrue(self.quota_checker.acquire_retry_quota(context))
        self.assertEqual(self.quota.available_capacity, 495)
        self.quota_checker.release_retry_quota(
            context.request_context, http_response=http_response
        )
        self.assertEqual(self.quota.available_capacity, 500)

    def test_dont_release_quota_if_all_retries_failed(self):
        context = self.create_context()
        # If max_attempts_reached is True, then it means we used up all
        # our retry attempts and still failed.  In this case we shouldn't
        # give any retry quota back.
        http_response = self.create_context(status_code=500).http_response
        self.assertTrue(self.quota_checker.acquire_retry_quota(context))
        self.assertEqual(self.quota.available_capacity, 486)
        self.quota_checker.release_retry_quota(
            context.request_context, http_response=http_response
        )
        self.assertEqual(self.quota.available_capacity, 486)

    def test_can_release_default_quota_if_not_in_context(self):
        context = self.create_context()
        self.assertTrue(self.quota_checker.acquire_retry_quota(context))
        self.assertEqual(self.quota.available_capacity, 486)
        # We're going to remove the quota amount from the request context.
        # This represents a successful request with no retries.
        self.request_context.pop('retry_quota_capacity')
        self.quota_checker.release_retry_quota(
            context.request_context, context.http_response
        )
        # We expect only 1 unit was released.
        self.assertEqual(self.quota.available_capacity, 487)

    def test_acquire_quota_fails(self):
        quota_checker = standard.RetryQuotaChecker(
            quota.RetryQuota(initial_capacity=14)
        )
        # The first one succeeds.
        self.assertTrue(
            quota_checker.acquire_retry_quota(self.create_context())
        )
        # But we should fail now because we're out of quota.
        self.request_context.pop('retry_quota_capacity')
        self.assertFalse(
            quota_checker.acquire_retry_quota(self.create_context())
        )
        self.assertNotIn('retry_quota_capacity', self.request_context)

    def test_quota_reached_adds_retry_metadata(self):
        quota_checker = standard.RetryQuotaChecker(
            quota.RetryQuota(initial_capacity=0)
        )
        context = self.create_context()
        self.assertFalse(quota_checker.acquire_retry_quota(context))
        self.assertEqual(
            context.get_retry_metadata(), {'RetryQuotaReached': True}
        )

    def test_single_failed_request_does_not_give_back_quota(self):
        context = self.create_context()
        http_response = self.create_context(status_code=400).http_response
        # First deduct some amount of the retry quota so we're not hitting
        # the upper bound.
        self.quota.acquire(50)
        self.assertEqual(self.quota.available_capacity, 450)
        self.quota_checker.release_retry_quota(
            context.request_context, http_response=http_response
        )
        self.assertEqual(self.quota.available_capacity, 450)


@mock.patch('botocore.retries.standard.NEW_RETRIES_ENABLED', True)
class TestServiceSpecificRetries(unittest.TestCase):
    def _make_retry_context(self, attempt, status_code, error_code=None):
        http_response = AWSResponse(
            status_code=status_code, raw=None, headers={}, url='https://foo/'
        )
        parsed_response = {}
        if error_code:
            parsed_response = {'Error': {'Code': error_code}}
        return standard.RetryContext(
            attempt_number=attempt,
            operation_model=mock.Mock(error_shapes=[]),
            http_response=http_response,
            parsed_response=parsed_response,
            request_context={},
        )

    def test_dynamodb_base_backoff_and_increased_retries(self):
        retry_quota_bucket = quota.RetryQuota()

        throttling_detector = standard.ThrottlingErrorDetector(
            standard.RetryEventAdapter()
        )
        retry_quota = standard.RetryQuotaChecker(
            retry_quota_bucket, throttling_detector
        )
        backoff = standard.ExponentialBackoff(
            random=lambda: 1,
            service_name='dynamodb',
            throttling_detector=throttling_detector,
        )

        retry_conditions = standard.StandardRetryConditions(max_attempts=4)

        # Attempts 1-3: retryable
        for attempt, expected_quota, expected_delay in [
            (1, 486, 0.025),
            (2, 472, 0.05),
            (3, 458, 0.1),
        ]:
            context = self._make_retry_context(
                attempt=attempt, status_code=500
            )
            self.assertTrue(retry_conditions.is_retryable(context))
            self.assertTrue(retry_quota.acquire_retry_quota(context))
            self.assertEqual(
                retry_quota_bucket.available_capacity, expected_quota
            )
            self.assertEqual(backoff.delay_amount(context), expected_delay)

        # Attempt 4: NOT retryable because max_attempts=4
        context4 = self._make_retry_context(attempt=4, status_code=500)
        self.assertFalse(retry_conditions.is_retryable(context4))
        self.assertEqual(retry_quota_bucket.available_capacity, 458)
        self.assertEqual(
            context4.get_retry_metadata(), {'MaxAttemptsReached': True}
        )

    def test_sqs_triggers_long_polling_backoff_when_token_empty(self):
        mock_sleep = mock.Mock()
        retry_policy = mock.Mock(spec=standard.RetryPolicy)
        retry_policy.should_retry.return_value = True
        retry_policy.compute_retry_delay.return_value = 0.05

        retry_quota = mock.Mock(spec=standard.RetryQuotaChecker)
        retry_quota.acquire_retry_quota.return_value = False

        handler = standard.RetryHandler(
            retry_policy=retry_policy,
            retry_event_adapter=standard.RetryEventAdapter(),
            retry_quota=retry_quota,
            service_name='sqs',
            sleep=mock_sleep,
        )

        context = self._make_retry_context(attempt=1, status_code=500)
        context.operation_model = mock.Mock()
        context.operation_model.name = 'ReceiveMessage'

        result = handler.needs_retry(
            response=(context.http_response, {}),
            attempts=1,
            caught_exception=None,
            request_dict={'context': {}},
            operation=context.operation_model,
        )

        assert result is False
        retry_quota.acquire_retry_quota.assert_called_once()
        mock_sleep.assert_called_once_with(0.05)

    def test_non_long_polling_operation_does_not_sleep_when_quota_exhausted(
        self,
    ):
        mock_sleep = mock.Mock()
        retry_policy = mock.Mock(spec=standard.RetryPolicy)
        retry_policy.should_retry.return_value = True
        retry_policy.compute_retry_delay.return_value = 0.05

        retry_quota = mock.Mock(spec=standard.RetryQuotaChecker)
        retry_quota.acquire_retry_quota.return_value = False

        handler = standard.RetryHandler(
            retry_policy=retry_policy,
            retry_event_adapter=standard.RetryEventAdapter(),
            retry_quota=retry_quota,
            service_name='sqs',
            sleep=mock_sleep,
        )

        context = self._make_retry_context(attempt=1, status_code=500)
        context.operation_model = mock.Mock()
        context.operation_model.name = 'SendMessage'

        result = handler.needs_retry(
            response=(context.http_response, {}),
            attempts=1,
            caught_exception=None,
            request_dict={'context': {}},
            operation=context.operation_model,
        )

        self.assertIsNone(result)
        mock_sleep.assert_not_called()


@mock.patch('botocore.retries.standard.NEW_RETRIES_ENABLED', True)
class TestRetryAfterHeaderInRetries:
    def setup_method(self):
        throttling_detector = standard.ThrottlingErrorDetector(
            standard.RetryEventAdapter()
        )
        self.retry_quota_bucket = quota.RetryQuota()
        self.retry_quota = standard.RetryQuotaChecker(
            self.retry_quota_bucket, throttling_detector
        )
        self.backoff = standard.ExponentialBackoff(
            random=lambda: 1,
            throttling_detector=throttling_detector,
        )

    def _make_retry_context(
        self, attempt, status_code, error_code=None, retry_after='0'
    ):
        http_response = AWSResponse(
            status_code=status_code,
            raw=None,
            headers={'x-amz-retry-after': retry_after},
            url='https://foo/',
        )
        parsed_response = {}
        if error_code:
            parsed_response = {'Error': {'Code': error_code}}
        return standard.RetryContext(
            attempt_number=attempt,
            operation_model=mock.Mock(error_shapes=[]),
            http_response=http_response,
            parsed_response=parsed_response,
            request_context={},
        )

    @pytest.mark.parametrize(
        'retry_after_header,'
        'expected_capacity_after_failure,'
        'expected_delay_amount,'
        'expected_capacity_after_success',
        [
            ('1500', 486, 1.5, 500),
            ('0', 486, 0.05, 500),
            ('10000', 486, 5.05, 500),
            ('invalid', 486, 0.05, 500),
            ('-100', 486, 0.05, 500),
        ],
    )
    def test_x_amz_retry_after_header_is_honored(
        self,
        retry_after_header,
        expected_capacity_after_failure,
        expected_delay_amount,
        expected_capacity_after_success,
    ):
        context = self._make_retry_context(
            attempt=1, status_code=500, retry_after=retry_after_header
        )

        assert self.retry_quota.acquire_retry_quota(context)
        assert (
            self.retry_quota_bucket.available_capacity
            == expected_capacity_after_failure
        )
        assert self.backoff.delay_amount(context) == expected_delay_amount

        http_success = AWSResponse(
            status_code=200, raw=None, headers={}, url='https://foo/'
        )
        self.retry_quota.release_retry_quota(
            context.request_context, http_response=http_success
        )
        assert (
            self.retry_quota_bucket.available_capacity
            == expected_capacity_after_success
        )


class TestNewRetriesEnvironmentVariable(BaseEnvVar):
    def test_env_var_true_enables_new_retries(self):
        self.environ['AWS_NEW_RETRIES_2026'] = 'true'
        self.assertTrue(configprovider._resolve_new_retries())

    def test_env_var_false_disables_new_retries(self):
        self.environ['AWS_NEW_RETRIES_2026'] = 'false'
        self.assertFalse(configprovider._resolve_new_retries())

    def test_no_env_var_uses_default(self):
        self.assertFalse(configprovider._resolve_new_retries())



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from unittest.mock import patch

from botocore.parsers import RestJSONParser
from botocore.session import Session
from botocore.utils import PRIORITY_ORDERED_SUPPORTED_PROTOCOLS
from tests import ClientHTTPStubber
from tests.functional import TEST_MODELS_DIR


def test_correct_protocol_selection():
    # This test ensures that botocore can properly resolve a protocol and
    # apply that to both the serializer and parser.  The logic for protocol
    # selection is deserialized for picking the serializer and parser, so
    # we add this test to make sure that in at least a simple circumstance
    # where a protocol is selected that is not the same as the "protocol"
    # trait, that we can properly create and use a serializer/parser.
    session = Session()
    loader = session.get_component('data_loader')
    loader.search_paths.insert(0, TEST_MODELS_DIR)

    client = session.create_client(
        'test-protocol-list',
        region_name='us-west-2',
        aws_access_key_id='foo',
        aws_secret_access_key='bar',
    )

    # This test would not be effective if the `protocol` trait was ever
    # a higher priority protocol than the resolved protocol as it would
    # pass even if a client's resolved parser was based on the `protocol`
    # trait instead of the resolved protocol
    service_model = client.meta.service_model
    assert PRIORITY_ORDERED_SUPPORTED_PROTOCOLS.index(
        service_model.protocol
    ) > PRIORITY_ORDERED_SUPPORTED_PROTOCOLS.index(
        service_model.resolved_protocol
    )

    original_do_parse = RestJSONParser._do_parse

    def tracking_do_parse(self, *args, **kwargs):
        called['was_called'] = True
        return original_do_parse(self, *args, **kwargs)

    with patch("botocore.parsers.RestJSONParser._do_parse", tracking_do_parse):
        called = {'was_called': False}
        with ClientHTTPStubber(client) as stubber:
            stubber.add_response(
                body=b'{"Bar": "Baz"}',
            )
            response = client.test_protocol_selection(Foo="input")
            assert response['Bar'] == 'Baz'
        assert called['was_called'], (
            "_do_parse was not called on RestJSONParser as expected"
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from botocore.session import get_session

_H2_REQUIRED = object()
# Service names to list of known HTTP 2 operations
_KNOWN_SERVICES = {
    'qbusiness': ['Chat'],
    'kinesis': ['SubscribeToShard'],
    'lexv2-runtime': ['StartConversation'],
    # Added only to keep a record of this feature being incompatible
    'bedrock-runtime': ['InvokeModelWithBidirectionalStream'],
    'connecthealth': ['StartMedicalScribeListeningSession'],
    'polly': ['StartSpeechSynthesisStream'],
}


def _all_test_cases():
    session = get_session()
    loader = session.get_component('data_loader')

    services = loader.list_available_services('service-2')
    h2_services = []
    h2_operations = []

    for service in services:
        service_model = session.get_service_model(service)
        h2_config = service_model.metadata.get('protocolSettings', {}).get(
            'h2'
        )
        if h2_config == 'required':
            h2_services.append(service)
        elif h2_config == 'eventstream':
            for operation in service_model.operation_names:
                operation_model = service_model.operation_model(operation)
                if operation_model.has_event_stream_output:
                    h2_operations.append([service, operation])

    return h2_services, h2_operations


H2_SERVICES, H2_OPERATIONS = _all_test_cases()


@pytest.mark.validates_models
@pytest.mark.parametrize("h2_service", H2_SERVICES)
def test_all_uses_of_h2_are_known(h2_service):
    # Validates that a service that requires HTTP 2 for all operations is known
    message = f'Found unknown HTTP 2 service: {h2_service}'
    assert _KNOWN_SERVICES.get(h2_service) is _H2_REQUIRED, message


@pytest.mark.validates_models
@pytest.mark.parametrize("h2_service, operation", H2_OPERATIONS)
def test_all_h2_operations_are_known(h2_service, operation):
    # Validates that an operation that requires HTTP 2 is known
    known_operations = _KNOWN_SERVICES.get(h2_service, [])
    message = f'Found unknown HTTP 2 operation: {h2_service}.{operation}'
    assert operation in known_operations, message



# Copyright 2026 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from botocore import xform_name
from botocore.session import get_session
from botocore.stub import Stubber

_OAUTH2_OPERATIONS = [
    'CreateOAuth2Token',
    'CreateOAuth2TokenWithIAM',
    'IntrospectOAuth2TokenWithIAM',
    'RevokeOAuth2TokenWithIAM',
]

_METHOD_RENAMES = [
    ('create_o_auth2_token_with_iam', 'create_oauth2_token_with_iam'),
    ('introspect_o_auth2_token_with_iam', 'introspect_oauth2_token_with_iam'),
    ('revoke_o_auth2_token_with_iam', 'revoke_oauth2_token_with_iam'),
]


@pytest.mark.validates_models
@pytest.mark.parametrize("operation", _OAUTH2_OPERATIONS)
def test_oauth2_xform_name(operation):
    assert 'oauth2' in xform_name(operation, '_')
    assert 'oauth2' in xform_name(operation, '-')


class TestSigninOAuth2:
    def setup_method(self):
        session = get_session()
        self.client = session.create_client('signin', 'us-east-1')
        self.stubber = Stubber(self.client)
        self.stubber.activate()

    def test_create_oauth2_token_aliased(self):
        # create_o_auth2_token was released without a rename, so we keep a
        # hidden alias for backwards compatibility.
        assert (
            self.client.create_o_auth2_token.__func__
            is self.client.create_oauth2_token.__func__
        )

    @pytest.mark.parametrize("modeled, renamed", _METHOD_RENAMES)
    def test_with_iam_methods_are_renamed(self, modeled, renamed):
        assert hasattr(self.client, renamed)
        assert not hasattr(self.client, modeled)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import os
from unittest import mock

import pytest
from dateutil.tz import tzutc

from botocore.config import Config
from tests import ClientHTTPStubber, temporary_file


class TestDisableS3ExpressAuth:
    DATE = datetime.datetime(2024, 11, 30, 23, 59, 59, tzinfo=tzutc())
    BUCKET_NAME = 'mybucket--usw2-az1--x-s3'

    CREATE_SESSION_RESPONSE = b'<?xml version="1.0" encoding="UTF-8"?>\n<CreateSessionResult><Credentials><AccessKeyId>test-key</AccessKeyId><Expiration>2024-12-31T23:59:59Z</Expiration><SecretAccessKey>test-secret</SecretAccessKey><SessionToken>test-token</SessionToken></Credentials></CreateSessionResult>'
    LIST_OBJECTS_RESPONSE = b'<?xml version="1.0" encoding="UTF-8"?>\n<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/"><Name>mybucket--usw2-az1--x-s3</Name><Prefix/><KeyCount>0</KeyCount><MaxKeys>1000</MaxKeys><EncodingType>url</EncodingType><IsTruncated>false</IsTruncated></ListBucketResult>'

    @pytest.fixture
    def mock_datetime(self):
        with mock.patch('datetime.datetime', spec=True) as mock_dt:
            mock_dt.now.return_value = self.DATE
            yield mock_dt

    def test_disable_s3_express_auth_enabled(
        self, patched_session, mock_datetime
    ):
        config = Config(s3={'disable_s3_express_session_auth': True})
        s3_client = patched_session.create_client(
            's3',
            config=config,
            region_name='us-west-2',
        )

        with ClientHTTPStubber(s3_client, strict=True) as stubber:
            stubber.add_response(body=self.LIST_OBJECTS_RESPONSE)
            s3_client.list_objects_v2(Bucket=self.BUCKET_NAME)

        assert len(stubber.requests) == 1

    def test_disable_s3_express_auth_enabled_env_var(
        self, patched_session, mock_datetime
    ):
        env = {'AWS_S3_DISABLE_EXPRESS_SESSION_AUTH': 'true'}
        with mock.patch.dict(os.environ, env):
            s3_client = patched_session.create_client(
                's3',
                region_name='us-west-2',
            )

            with ClientHTTPStubber(s3_client, strict=True) as stubber:
                stubber.add_response(body=self.LIST_OBJECTS_RESPONSE)
                s3_client.list_objects_v2(Bucket=self.BUCKET_NAME)

            assert len(stubber.requests) == 1

    def test_disable_s3_express_auth_enabled_shared_config(
        self, patched_session, mock_datetime
    ):
        with temporary_file('w') as f:
            f.write('[default]\n')
            f.write('s3_disable_express_session_auth = true\n')
            f.flush()

            with mock.patch.dict(os.environ, {'AWS_CONFIG_FILE': f.name}):
                s3_client = patched_session.create_client(
                    's3',
                    region_name='us-west-2',
                )

                with ClientHTTPStubber(s3_client, strict=True) as stubber:
                    stubber.add_response(body=self.LIST_OBJECTS_RESPONSE)
                    s3_client.list_objects_v2(Bucket=self.BUCKET_NAME)

                assert len(stubber.requests) == 1

    def test_disable_s3_express_auth_disabled(
        self, patched_session, mock_datetime
    ):
        config = Config(s3={'disable_s3_express_session_auth': False})
        s3_client = patched_session.create_client(
            's3',
            config=config,
            region_name='us-west-2',
        )

        with ClientHTTPStubber(s3_client, strict=True) as stubber:
            stubber.add_response(body=self.CREATE_SESSION_RESPONSE)
            stubber.add_response(body=self.LIST_OBJECTS_RESPONSE)
            s3_client.list_objects_v2(Bucket=self.BUCKET_NAME)

        assert len(stubber.requests) == 2



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import importlib
import os
import sys
from unittest import mock

import pytest

from botocore.plugin import (
    PluginContext,
    reset_plugin_context,
    set_plugin_context,
)
from botocore.session import get_session
from tests import ClientHTTPStubber


@pytest.fixture(scope="module", autouse=True)
def sys_mock():
    _preserved_sys_path = sys.path[:]
    plugins_dir = os.path.join(os.path.dirname(__file__), "plugins")
    sys.path.append(plugins_dir)
    try:
        yield
    finally:
        sys.path = _preserved_sys_path


def mock_env_for_plugins(plugins_value):
    return mock.patch.dict(
        os.environ,
        {
            'AWS_ACCESS_KEY_ID': 'access_key',
            "AWS_SECRET_ACCESS_KEY": "secret_key",
            'BOTOCORE_EXPERIMENTAL__PLUGINS': plugins_value,
        },
    )


def test_environment_variable():
    plugin_module = importlib.import_module("simple_plugin_test_module")
    test_plugin = plugin_module.plugin_instance
    with mock_env_for_plugins('plugin_name=simple_plugin_test_module'):
        session = get_session()
        client = session.create_client('dynamodb', region_name='us-east-1')
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=200, body=b'')
            client.list_tables()
        assert len(test_plugin.events_seen) == 1
        assert isinstance(test_plugin.events_seen[0], dict)


def test_recursive_plugin_module():
    plugin_module = importlib.import_module("recursive_plugin_test_module")
    recursive_plugin = plugin_module.plugin_instance
    with mock_env_for_plugins('recursive=recursive_plugin_test_module'):
        session = get_session()
        session.set_credentials('key', 'secret')
        client = session.create_client('dynamodb', region_name='us-east-1')
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=200, body=b'')
            client.list_tables()
    assert recursive_plugin.called


def test_plugin_not_loaded_when_disabled():
    plugin_mod = importlib.import_module("disabled_test_plugin")
    disabled_plugin = plugin_mod.plugin_instance
    with mock_env_for_plugins('disabled_plugin=disabled_test_plugin'):
        ctx = PluginContext(plugins="DISABLED")
        token = set_plugin_context(ctx)
        try:
            session = get_session()
            session.set_credentials('key', 'secret')
            client = session.create_client('dynamodb', region_name='us-east-1')
            with ClientHTTPStubber(client) as http_stubber:
                http_stubber.add_response(status=200, body=b'')
                client.list_tables()
        finally:
            reset_plugin_context(token)
    assert disabled_plugin.invocations == 0


def test_multiple_plugins_and_malformed():
    plugin1_mod = importlib.import_module("first_of_two_plugins")
    plugin2_mod = importlib.import_module("second_of_two_plugins")
    first_plugin = plugin1_mod.plugin_instance
    second_plugin = plugin2_mod.plugin_instance
    with mock_env_for_plugins(
        'plugin1=first_of_two_plugins,plugin2=second_of_two_plugins,malformedplugin'
    ):
        session = get_session()
        session.set_credentials('key', 'secret')
        client = session.create_client('dynamodb', region_name='us-east-1')
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=200, body=b'')
            client.list_tables()
    assert first_plugin.invocations == 1
    assert second_plugin.invocations == 1



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import json

import pytest
from dateutil.tz import tzutc

import botocore
import botocore.auth
import botocore.awsrequest
import botocore.session
from botocore.config import Config
from botocore.credentials import Credentials, ReadOnlyCredentials
from botocore.exceptions import (
    NoRegionError,
    ParamValidationError,
    UnknownClientMethodError,
    UnknownSignatureVersionError,
    UnsupportedSignatureVersionError,
)
from botocore.hooks import HierarchicalEmitter
from botocore.model import ServiceId
from botocore.signers import (
    CloudFrontSigner,
    RequestSigner,
    S3PostPresigner,
    _dsql_generate_db_auth_token,
    dsql_generate_db_connect_admin_auth_token,
    dsql_generate_db_connect_auth_token,
    generate_db_auth_token,
)
from tests import FreezeTime, assert_url_equal, mock, unittest

DATE = datetime.datetime(2024, 11, 7, 17, 39, 33, tzinfo=tzutc())


@pytest.fixture
def polly_client():
    session = botocore.session.get_session()
    session.set_credentials('key', 'secret')
    return session.create_client('polly', region_name='us-west-2')


class BaseSignerTest(unittest.TestCase):
    def setUp(self):
        self.credentials = Credentials('key', 'secret')
        self.emitter = mock.Mock()
        self.emitter.emit_until_response.return_value = (None, None)
        self.signer = RequestSigner(
            ServiceId('service_name'),
            'region_name',
            'signing_name',
            'v4',
            self.credentials,
            self.emitter,
        )
        self.fixed_credentials = self.credentials.get_frozen_credentials()
        self.request = botocore.awsrequest.AWSRequest()


class TestSigner(BaseSignerTest):
    def test_region_name(self):
        self.assertEqual(self.signer.region_name, 'region_name')

    def test_signature_version(self):
        self.assertEqual(self.signer.signature_version, 'v4')

    def test_signing_name(self):
        self.assertEqual(self.signer.signing_name, 'signing_name')

    def test_region_required_for_sigv4(self):
        self.signer = RequestSigner(
            ServiceId('service_name'),
            None,
            'signing_name',
            'v4',
            self.credentials,
            self.emitter,
        )

        with self.assertRaises(NoRegionError):
            self.signer.sign('operation_name', self.request)

    def test_get_auth(self):
        auth_cls = mock.Mock()
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4': auth_cls}):
            auth = self.signer.get_auth('service_name', 'region_name')

            self.assertEqual(auth, auth_cls.return_value)
            auth_cls.assert_called_with(
                credentials=self.fixed_credentials,
                service_name='service_name',
                region_name='region_name',
            )

    def test_get_auth_signature_override(self):
        auth_cls = mock.Mock()
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'v4-custom': auth_cls}
        ):
            auth = self.signer.get_auth(
                'service_name', 'region_name', signature_version='v4-custom'
            )

            self.assertEqual(auth, auth_cls.return_value)
            auth_cls.assert_called_with(
                credentials=self.fixed_credentials,
                service_name='service_name',
                region_name='region_name',
            )

    def test_get_auth_bad_override(self):
        with self.assertRaises(UnknownSignatureVersionError):
            self.signer.get_auth(
                'service_name', 'region_name', signature_version='bad'
            )

    def test_emits_choose_signer(self):
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'v4': mock.Mock()}
        ):
            self.signer.sign('operation_name', self.request)

        self.emitter.emit_until_response.assert_called_with(
            'choose-signer.service_name.operation_name',
            signing_name='signing_name',
            region_name='region_name',
            signature_version='v4',
            context=mock.ANY,
        )

    def test_choose_signer_override(self):
        auth = mock.Mock()
        auth.REQUIRES_REGION = False
        self.emitter.emit_until_response.return_value = (None, 'custom')

        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'custom': auth}):
            self.signer.sign('operation_name', self.request)

        auth.assert_called_with(credentials=self.fixed_credentials)
        auth.return_value.add_auth.assert_called_with(self.request)

    def test_emits_before_sign(self):
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'v4': mock.Mock()}
        ):
            self.signer.sign('operation_name', self.request)

        self.emitter.emit.assert_called_with(
            'before-sign.service_name.operation_name',
            request=self.request,
            signing_name='signing_name',
            region_name='region_name',
            signature_version='v4',
            request_signer=self.signer,
            operation_name='operation_name',
        )

    def test_disable_signing(self):
        # Returning botocore.UNSIGNED from choose-signer disables signing!
        auth = mock.Mock()
        self.emitter.emit_until_response.return_value = (
            None,
            botocore.UNSIGNED,
        )

        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4': auth}):
            self.signer.sign('operation_name', self.request)

        auth.assert_not_called()

    def test_generate_url_emits_choose_signer(self):
        request_dict = {
            'headers': {},
            'url': 'https://foo.com',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }

        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'v4': mock.Mock()}
        ):
            self.signer.generate_presigned_url(request_dict, 'operation_name')

        self.emitter.emit_until_response.assert_called_with(
            'choose-signer.service_name.operation_name',
            signing_name='signing_name',
            region_name='region_name',
            signature_version='v4-query',
            context=mock.ANY,
        )

    def test_choose_signer_passes_context(self):
        self.request.context = {'foo': 'bar'}

        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'v4': mock.Mock()}
        ):
            self.signer.sign('operation_name', self.request)

        self.emitter.emit_until_response.assert_called_with(
            'choose-signer.service_name.operation_name',
            signing_name='signing_name',
            region_name='region_name',
            signature_version='v4',
            context={'foo': 'bar'},
        )

    def test_generate_url_choose_signer_override(self):
        request_dict = {
            'headers': {},
            'url': 'https://foo.com',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }
        auth = mock.Mock()
        auth.REQUIRES_REGION = False
        self.emitter.emit_until_response.return_value = (None, 'custom')

        auth_types_map = {'custom': mock.Mock(), 'custom-query': auth}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types_map):
            self.signer.generate_presigned_url(request_dict, 'operation_name')

        auth.assert_called_with(
            credentials=self.fixed_credentials, expires=3600
        )

    def test_generate_url_unsigned(self):
        request_dict = {
            'headers': {},
            'url': 'https://foo.com',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }
        self.emitter.emit_until_response.return_value = (
            None,
            botocore.UNSIGNED,
        )

        url = self.signer.generate_presigned_url(
            request_dict, 'operation_name'
        )

        self.assertEqual(url, 'https://foo.com')

    def test_generate_presigned_url(self):
        auth = mock.Mock()
        auth.REQUIRES_REGION = True

        request_dict = {
            'headers': {},
            'url': 'https://foo.com',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4-query': auth}):
            presigned_url = self.signer.generate_presigned_url(
                request_dict, operation_name='operation_name'
            )
        auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='region_name',
            service_name='signing_name',
            expires=3600,
        )
        self.assertEqual(presigned_url, 'https://foo.com')

    def test_generate_presigned_url_with_region_override(self):
        auth = mock.Mock()
        auth.REQUIRES_REGION = True

        request_dict = {
            'headers': {},
            'url': 'https://foo.com',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4-query': auth}):
            presigned_url = self.signer.generate_presigned_url(
                request_dict,
                operation_name='operation_name',
                region_name='us-west-2',
            )
        auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='us-west-2',
            service_name='signing_name',
            expires=3600,
        )
        self.assertEqual(presigned_url, 'https://foo.com')

    def test_generate_presigned_url_with_exipres_in(self):
        auth = mock.Mock()
        auth.REQUIRES_REGION = True

        request_dict = {
            'headers': {},
            'url': 'https://foo.com',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4-query': auth}):
            presigned_url = self.signer.generate_presigned_url(
                request_dict, operation_name='operation_name', expires_in=900
            )
        auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='region_name',
            expires=900,
            service_name='signing_name',
        )
        self.assertEqual(presigned_url, 'https://foo.com')

    def test_presigned_url_throws_unsupported_signature_error(self):
        request_dict = {
            'headers': {},
            'url': 'https://s3.amazonaws.com/mybucket/myobject',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }
        self.signer = RequestSigner(
            ServiceId('service_name'),
            'region_name',
            'signing_name',
            'foo',
            self.credentials,
            self.emitter,
        )
        with self.assertRaises(UnsupportedSignatureVersionError):
            self.signer.generate_presigned_url(
                request_dict, operation_name='foo'
            )

    def test_signer_with_refreshable_credentials_gets_credential_set(self):
        class FakeCredentials(Credentials):
            def get_frozen_credentials(self):
                return ReadOnlyCredentials('foo', 'bar', 'baz')

        self.credentials = FakeCredentials('a', 'b', 'c')

        self.signer = RequestSigner(
            ServiceId('service_name'),
            'region_name',
            'signing_name',
            'v4',
            self.credentials,
            self.emitter,
        )

        auth_cls = mock.Mock()
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4': auth_cls}):
            auth = self.signer.get_auth('service_name', 'region_name')
            self.assertEqual(auth, auth_cls.return_value)
            # Note we're called with 'foo', 'bar', 'baz', and *not*
            # 'a', 'b', 'c'.
            auth_cls.assert_called_with(
                credentials=ReadOnlyCredentials('foo', 'bar', 'baz'),
                service_name='service_name',
                region_name='region_name',
            )

    def test_no_credentials_case_is_forwarded_to_signer(self):
        # If no credentials are given to the RequestSigner, we should
        # forward that fact on to the Auth class and let them handle
        # the error (which they already do).
        self.credentials = None
        self.signer = RequestSigner(
            ServiceId('service_name'),
            'region_name',
            'signing_name',
            'v4',
            self.credentials,
            self.emitter,
        )
        auth_cls = mock.Mock()
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4': auth_cls}):
            self.signer.get_auth_instance('service_name', 'region_name', 'v4')
            auth_cls.assert_called_with(
                service_name='service_name',
                region_name='region_name',
                credentials=None,
            )

    def test_sign_with_signing_type_standard(self):
        auth = mock.Mock()
        post_auth = mock.Mock()
        query_auth = mock.Mock()
        auth_types = {
            'v4-presign-post': post_auth,
            'v4-query': query_auth,
            'v4': auth,
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign(
                'operation_name', self.request, signing_type='standard'
            )
        self.assertFalse(post_auth.called)
        self.assertFalse(query_auth.called)
        auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='signing_name',
            region_name='region_name',
        )

    def test_sign_with_signing_type_presign_url(self):
        auth = mock.Mock()
        post_auth = mock.Mock()
        query_auth = mock.Mock()
        auth_types = {
            'v4-presign-post': post_auth,
            'v4-query': query_auth,
            'v4': auth,
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign(
                'operation_name', self.request, signing_type='presign-url'
            )
        self.assertFalse(post_auth.called)
        self.assertFalse(auth.called)
        query_auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='signing_name',
            region_name='region_name',
        )

    def test_sign_with_signing_type_presign_post(self):
        auth = mock.Mock()
        post_auth = mock.Mock()
        query_auth = mock.Mock()
        auth_types = {
            'v4-presign-post': post_auth,
            'v4-query': query_auth,
            'v4': auth,
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign(
                'operation_name', self.request, signing_type='presign-post'
            )
        self.assertFalse(auth.called)
        self.assertFalse(query_auth.called)
        post_auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='signing_name',
            region_name='region_name',
        )

    def test_sign_with_region_name(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign('operation_name', self.request, region_name='foo')
        auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='signing_name',
            region_name='foo',
        )

    def test_sign_override_region_from_context(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        self.request.context = {'signing': {'region': 'my-override-region'}}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign('operation_name', self.request)
        auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='signing_name',
            region_name='my-override-region',
        )

    def test_sign_with_region_name_overrides_context(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        self.request.context = {'signing': {'region': 'context-override'}}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign(
                'operation_name', self.request, region_name='param-override'
            )
        auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='signing_name',
            region_name='param-override',
        )

    def test_sign_override_signing_name_from_context(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        self.request.context = {'signing': {'signing_name': 'override_name'}}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign('operation_name', self.request)
        auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='override_name',
            region_name='region_name',
        )

    def test_sign_with_expires_in(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign('operation_name', self.request, expires_in=2)
        auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='signing_name',
            region_name='region_name',
            expires=2,
        )

    def test_sign_with_custom_signing_name(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.sign(
                'operation_name', self.request, signing_name='foo'
            )
        auth.assert_called_with(
            credentials=ReadOnlyCredentials('key', 'secret', None),
            service_name='foo',
            region_name='region_name',
        )

    def test_presign_with_custom_signing_name(self):
        auth = mock.Mock()
        auth.REQUIRES_REGION = True

        request_dict = {
            'headers': {},
            'url': 'https://foo.com',
            'body': b'',
            'url_path': '/',
            'method': 'GET',
            'context': {},
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, {'v4-query': auth}):
            presigned_url = self.signer.generate_presigned_url(
                request_dict,
                operation_name='operation_name',
                signing_name='foo',
            )
        auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='region_name',
            expires=3600,
            service_name='foo',
        )
        self.assertEqual(presigned_url, 'https://foo.com')

    def test_unknown_signer_raises_unknown_on_standard(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        self.emitter.emit_until_response.return_value = (None, 'custom')
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            with self.assertRaises(UnknownSignatureVersionError):
                self.signer.sign(
                    'operation_name', self.request, signing_type='standard'
                )

    def test_unknown_signer_raises_unsupported_when_not_standard(self):
        auth = mock.Mock()
        auth_types = {'v4': auth}
        self.emitter.emit_until_response.return_value = (None, 'custom')
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            with self.assertRaises(UnsupportedSignatureVersionError):
                self.signer.sign(
                    'operation_name', self.request, signing_type='presign-url'
                )

            with self.assertRaises(UnsupportedSignatureVersionError):
                self.signer.sign(
                    'operation_name', self.request, signing_type='presign-post'
                )


class TestCloudfrontSigner(BaseSignerTest):
    def setUp(self):
        super().setUp()
        self.signer = CloudFrontSigner("MY_KEY_ID", lambda message: b'signed')
        # It helps but the long string diff will still be slightly different on
        # Python 2.6/2.7/3.x. We won't soly rely on that anyway, so it's fine.
        self.maxDiff = None

    def test_build_canned_policy(self):
        policy = self.signer.build_policy('foo', datetime.datetime(2016, 1, 1))
        expected = (
            '{"Statement":[{"Resource":"foo",'
            '"Condition":{"DateLessThan":{"AWS:EpochTime":1451606400}}}]}'
        )
        self.assertEqual(json.loads(policy), json.loads(expected))
        self.assertEqual(policy, expected)  # This is to ensure the right order

    def test_build_custom_policy(self):
        policy = self.signer.build_policy(
            'foo',
            datetime.datetime(2016, 1, 1),
            date_greater_than=datetime.datetime(2015, 12, 1),
            ip_address='12.34.56.78/9',
        )
        expected = {
            "Statement": [
                {
                    "Resource": "foo",
                    "Condition": {
                        "DateGreaterThan": {"AWS:EpochTime": 1448928000},
                        "DateLessThan": {"AWS:EpochTime": 1451606400},
                        "IpAddress": {"AWS:SourceIp": "12.34.56.78/9"},
                    },
                }
            ]
        }
        self.assertEqual(json.loads(policy), expected)

    def test_generate_presign_url_with_expire_time(self):
        signed_url = self.signer.generate_presigned_url(
            'http://test.com/foo.txt',
            date_less_than=datetime.datetime(2016, 1, 1),
        )
        expected = (
            'http://test.com/foo.txt?Expires=1451606400&Signature=c2lnbmVk'
            '&Key-Pair-Id=MY_KEY_ID'
        )
        assert_url_equal(signed_url, expected)

    def test_generate_presign_url_with_custom_policy(self):
        policy = self.signer.build_policy(
            'foo',
            datetime.datetime(2016, 1, 1),
            date_greater_than=datetime.datetime(2015, 12, 1),
            ip_address='12.34.56.78/9',
        )
        signed_url = self.signer.generate_presigned_url(
            'http://test.com/index.html?foo=bar', policy=policy
        )
        expected = (
            'http://test.com/index.html?foo=bar'
            '&Policy=eyJTdGF0ZW1lbnQiOlt7IlJlc291cmNlIjoiZm9vIiwiQ29uZ'
            'Gl0aW9uIjp7IkRhdGVMZXNzVGhhbiI6eyJBV1M6RXBvY2hUaW1lIj'
            'oxNDUxNjA2NDAwfSwiSXBBZGRyZXNzIjp7IkFXUzpTb3VyY2VJcCI'
            '6IjEyLjM0LjU2Ljc4LzkifSwiRGF0ZUdyZWF0ZXJUaGFuIjp7IkFX'
            'UzpFcG9jaFRpbWUiOjE0NDg5MjgwMDB9fX1dfQ__'
            '&Signature=c2lnbmVk&Key-Pair-Id=MY_KEY_ID'
        )
        assert_url_equal(signed_url, expected)


class TestS3PostPresigner(BaseSignerTest):
    def setUp(self):
        super().setUp()
        self.request_signer = RequestSigner(
            ServiceId('service_name'),
            'region_name',
            'signing_name',
            's3v4',
            self.credentials,
            self.emitter,
        )
        self.signer = S3PostPresigner(self.request_signer)
        self.request_dict = {
            'headers': {},
            'url': 'https://s3.amazonaws.com/mybucket',
            'body': b'',
            'url_path': '/',
            'method': 'POST',
            'context': {},
        }
        self.auth = mock.Mock()
        self.auth.REQUIRES_REGION = True
        self.add_auth = mock.Mock()
        self.auth.return_value.add_auth = self.add_auth
        self.fixed_credentials = self.credentials.get_frozen_credentials()

        self.datetime_patch = mock.patch('botocore.compat.datetime')
        self.datetime_mock = self.datetime_patch.start()
        self.fixed_date = datetime.datetime(2014, 3, 10, 17, 2, 55, 0)
        self.fixed_delta = datetime.timedelta(seconds=3600)
        self.datetime_mock.datetime.now.return_value = self.fixed_date
        self.datetime_mock.timedelta.return_value = self.fixed_delta

    def tearDown(self):
        super().tearDown()
        self.datetime_patch.stop()

    def test_generate_presigned_post(self):
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'s3v4-presign-post': self.auth}
        ):
            post_form_args = self.signer.generate_presigned_post(
                self.request_dict
            )
        self.auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='region_name',
            service_name='signing_name',
        )
        self.assertEqual(self.add_auth.call_count, 1)
        ref_request = self.add_auth.call_args[0][0]
        ref_policy = ref_request.context['s3-presign-post-policy']
        self.assertEqual(ref_policy['expiration'], '2014-03-10T18:02:55Z')
        self.assertEqual(ref_policy['conditions'], [])

        self.assertEqual(
            post_form_args['url'], 'https://s3.amazonaws.com/mybucket'
        )
        self.assertEqual(post_form_args['fields'], {})

    def test_generate_presigned_post_emits_choose_signer(self):
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'s3v4-presign-post': self.auth}
        ):
            self.signer.generate_presigned_post(self.request_dict)
        self.emitter.emit_until_response.assert_called_with(
            'choose-signer.service_name.PutObject',
            signing_name='signing_name',
            region_name='region_name',
            signature_version='s3v4-presign-post',
            context=mock.ANY,
        )

    def test_generate_presigned_post_choose_signer_override(self):
        auth = mock.Mock()
        self.emitter.emit_until_response.return_value = (None, 'custom')
        auth_types = {
            's3v4-presign-post': self.auth,
            'custom-presign-post': auth,
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.generate_presigned_post(self.request_dict)
        auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='region_name',
            service_name='signing_name',
        )

    def test_generate_presigne_post_choose_signer_override_known(self):
        auth = mock.Mock()
        self.emitter.emit_until_response.return_value = (
            None,
            's3v4-presign-post',
        )
        auth_types = {
            's3v4-presign-post': self.auth,
            'custom-presign-post': auth,
        }
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            self.signer.generate_presigned_post(self.request_dict)
        self.auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='region_name',
            service_name='signing_name',
        )

    def test_generate_presigned_post_bad_signer_raises_error(self):
        auth = mock.Mock()
        self.emitter.emit_until_response.return_value = (None, 's3-query')
        auth_types = {'s3v4-presign-post': self.auth, 's3-query': auth}
        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
            with self.assertRaises(UnsupportedSignatureVersionError):
                self.signer.generate_presigned_post(self.request_dict)

    def test_generate_unsigned_post(self):
        self.emitter.emit_until_response.return_value = (
            None,
            botocore.UNSIGNED,
        )
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'s3v4-presign-post': self.auth}
        ):
            post_form_args = self.signer.generate_presigned_post(
                self.request_dict
            )
        expected = {'fields': {}, 'url': 'https://s3.amazonaws.com/mybucket'}
        self.assertEqual(post_form_args, expected)

    def test_generate_presigned_post_with_conditions(self):
        conditions = [{'bucket': 'mybucket'}, ['starts-with', '$key', 'bar']]
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'s3v4-presign-post': self.auth}
        ):
            self.signer.generate_presigned_post(
                self.request_dict, conditions=conditions
            )
        self.auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='region_name',
            service_name='signing_name',
        )
        self.assertEqual(self.add_auth.call_count, 1)
        ref_request = self.add_auth.call_args[0][0]
        ref_policy = ref_request.context['s3-presign-post-policy']
        self.assertEqual(ref_policy['conditions'], conditions)

    def test_generate_presigned_post_with_region_override(self):
        with mock.patch.dict(
            botocore.auth.AUTH_TYPE_MAPS, {'s3v4-presign-post': self.auth}
        ):
            self.signer.generate_presigned_post(
                self.request_dict, region_name='foo'
            )
        self.auth.assert_called_with(
            credentials=self.fixed_credentials,
            region_name='foo',
            service_name='signing_name',
        )

    def test_presigned_post_throws_unsupported_signature_error(self):
        request_dict = {
            'headers': {},
            'url': 'https://s3.amazonaws.com/mybucket/myobject',
            'body': b'',
            'url_path': '/',
            'method': 'POST',
            'context': {},
        }
        self.request_signer = RequestSigner(
            ServiceId('service_name'),
            'region_name',
            'signing_name',
            'foo',
            self.credentials,
            self.emitter,
        )
        self.signer = S3PostPresigner(self.request_signer)
        with self.assertRaises(UnsupportedSignatureVersionError):
            self.signer.generate_presigned_post(request_dict)


class TestGenerateUrl(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client('s3', region_name='us-east-1')
        self.bucket = 'mybucket'
        self.key = 'mykey'
        self.client_kwargs = {'Bucket': self.bucket, 'Key': self.key}
        self.generate_url_patch = mock.patch(
            'botocore.signers.RequestSigner.generate_presigned_url'
        )
        self.generate_url_mock = self.generate_url_patch.start()

    def tearDown(self):
        self.generate_url_patch.stop()

    def test_generate_presigned_url(self):
        self.client.generate_presigned_url(
            'get_object', Params={'Bucket': self.bucket, 'Key': self.key}
        )

        ref_request_dict = {
            'body': b'',
            'url': 'https://mybucket.s3.amazonaws.com/mykey',
            'headers': {},
            'auth_path': '/mybucket/mykey',
            'query_string': {},
            'url_path': '/mykey',
            'method': 'GET',
            # mock.ANY is used because client parameter related events
            # inject values into the context. So using the context's exact
            # value for these tests will be a maintenance burden if
            # anymore customizations are added that inject into the context.
            'context': mock.ANY,
        }
        self.generate_url_mock.assert_called_with(
            request_dict=ref_request_dict,
            expires_in=3600,
            operation_name='GetObject',
        )

    def test_generate_presigned_url_with_query_string(self):
        disposition = 'attachment; filename="download.jpg"'
        self.client.generate_presigned_url(
            'get_object',
            Params={
                'Bucket': self.bucket,
                'Key': self.key,
                'ResponseContentDisposition': disposition,
            },
        )
        ref_request_dict = {
            'body': b'',
            'url': (
                'https://mybucket.s3.amazonaws.com/mykey'
                '?response-content-disposition='
                'attachment%3B%20filename%3D%22download.jpg%22'
            ),
            'auth_path': '/mybucket/mykey',
            'headers': {},
            'query_string': {'response-content-disposition': disposition},
            'url_path': '/mykey',
            'method': 'GET',
            'context': mock.ANY,
        }
        self.generate_url_mock.assert_called_with(
            request_dict=ref_request_dict,
            expires_in=3600,
            operation_name='GetObject',
        )

    def test_generate_presigned_url_unknown_method_name(self):
        with self.assertRaises(UnknownClientMethodError):
            self.client.generate_presigned_url('getobject')

    def test_generate_presigned_url_missing_required_params(self):
        with self.assertRaises(ParamValidationError):
            self.client.generate_presigned_url('get_object')

    def test_generate_presigned_url_expires(self):
        self.client.generate_presigned_url(
            'get_object',
            Params={'Bucket': self.bucket, 'Key': self.key},
            ExpiresIn=20,
        )
        ref_request_dict = {
            'body': b'',
            'url': 'https://mybucket.s3.amazonaws.com/mykey',
            'auth_path': '/mybucket/mykey',
            'headers': {},
            'query_string': {},
            'url_path': '/mykey',
            'method': 'GET',
            'context': mock.ANY,
        }
        self.generate_url_mock.assert_called_with(
            request_dict=ref_request_dict,
            expires_in=20,
            operation_name='GetObject',
        )

    def test_generate_presigned_url_override_http_method(self):
        self.client.generate_presigned_url(
            'get_object',
            Params={'Bucket': self.bucket, 'Key': self.key},
            HttpMethod='PUT',
        )
        ref_request_dict = {
            'body': b'',
            'url': 'https://mybucket.s3.amazonaws.com/mykey',
            'auth_path': '/mybucket/mykey',
            'headers': {},
            'query_string': {},
            'url_path': '/mykey',
            'method': 'PUT',
            'context': mock.ANY,
        }
        self.generate_url_mock.assert_called_with(
            request_dict=ref_request_dict,
            expires_in=3600,
            operation_name='GetObject',
        )

    def test_generate_presigned_url_emits_param_events(self):
        emitter = mock.Mock(HierarchicalEmitter)
        emitter.emit.return_value = []
        self.client.meta.events = emitter
        self.client.generate_presigned_url(
            'get_object', Params={'Bucket': self.bucket, 'Key': self.key}
        )
        events_emitted = [
            emit_call[0][0] for emit_call in emitter.emit.call_args_list
        ]
        self.assertEqual(
            events_emitted,
            [
                'provide-client-params.s3.GetObject',
                'before-parameter-build.s3.GetObject',
            ],
        )

    def test_generate_presign_url_emits_is_presign_in_context(self):
        emitter = mock.Mock(HierarchicalEmitter)
        emitter.emit.return_value = []
        self.client.meta.events = emitter
        self.client.generate_presigned_url(
            'get_object', Params={'Bucket': self.bucket, 'Key': self.key}
        )
        kwargs_emitted = [
            emit_call[1] for emit_call in emitter.emit.call_args_list
        ]
        for kwargs in kwargs_emitted:
            self.assertTrue(
                kwargs.get('context', {}).get('is_presign_request'),
                'The context did not have is_presign_request set to True for '
                f'the following kwargs emitted: {kwargs}',
            )

    def test_context_param_from_event_handler_sent_to_endpoint_resolver(self):
        def change_bucket_param(params, **kwargs):
            params['Bucket'] = 'mybucket-bar'

        self.client.meta.events.register_last(
            'provide-client-params.s3.*', change_bucket_param
        )

        self.client.generate_presigned_url(
            'get_object', Params={'Bucket': 'mybucket-foo', 'Key': self.key}
        )

        ref_request_dict = {
            'body': b'',
            # If the bucket name set in the provide-client-params event handler
            # was correctly passed to the endpoint provider as a dynamic context
            # parameter, it will appear in the URL and the auth_path:
            'url': 'https://mybucket-bar.s3.amazonaws.com/mykey',
            'headers': {},
            'auth_path': '/mybucket-bar/mykey',
            'query_string': {},
            'url_path': '/mykey',
            'method': 'GET',
            'context': mock.ANY,
        }
        self.generate_url_mock.assert_called_with(
            request_dict=ref_request_dict,
            expires_in=3600,
            operation_name='GetObject',
        )


class TestGeneratePresignedPost(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client('s3', region_name='us-east-1')
        self.bucket = 'mybucket'
        self.key = 'mykey'
        self.presign_post_patch = mock.patch(
            'botocore.signers.S3PostPresigner.generate_presigned_post'
        )
        self.presign_post_mock = self.presign_post_patch.start()

    def tearDown(self):
        self.presign_post_patch.stop()

    def test_generate_presigned_post(self):
        self.client.generate_presigned_post(self.bucket, self.key)

        _, post_kwargs = self.presign_post_mock.call_args
        request_dict = post_kwargs['request_dict']
        fields = post_kwargs['fields']
        conditions = post_kwargs['conditions']
        self.assertEqual(
            request_dict['url'], 'https://mybucket.s3.amazonaws.com/'
        )
        self.assertEqual(post_kwargs['expires_in'], 3600)
        self.assertEqual(
            conditions, [{'bucket': 'mybucket'}, {'key': 'mykey'}]
        )
        self.assertEqual(fields, {'key': 'mykey'})

    def test_generate_presigned_post_with_filename(self):
        self.key = 'myprefix/${filename}'
        self.client.generate_presigned_post(self.bucket, self.key)

        _, post_kwargs = self.presign_post_mock.call_args
        request_dict = post_kwargs['request_dict']
        fields = post_kwargs['fields']
        conditions = post_kwargs['conditions']
        self.assertEqual(
            request_dict['url'], 'https://mybucket.s3.amazonaws.com/'
        )
        self.assertEqual(post_kwargs['expires_in'], 3600)
        self.assertEqual(
            conditions,
            [{'bucket': 'mybucket'}, ['starts-with', '$key', 'myprefix/']],
        )
        self.assertEqual(fields, {'key': 'myprefix/${filename}'})

    def test_generate_presigned_post_expires(self):
        self.client.generate_presigned_post(
            self.bucket, self.key, ExpiresIn=50
        )
        _, post_kwargs = self.presign_post_mock.call_args
        request_dict = post_kwargs['request_dict']
        fields = post_kwargs['fields']
        conditions = post_kwargs['conditions']
        self.assertEqual(
            request_dict['url'], 'https://mybucket.s3.amazonaws.com/'
        )
        self.assertEqual(post_kwargs['expires_in'], 50)
        self.assertEqual(
            conditions, [{'bucket': 'mybucket'}, {'key': 'mykey'}]
        )
        self.assertEqual(fields, {'key': 'mykey'})

    def test_generate_presigned_post_with_prefilled(self):
        conditions = [{'acl': 'public-read'}]
        fields = {'acl': 'public-read'}

        self.client.generate_presigned_post(
            self.bucket, self.key, Fields=fields, Conditions=conditions
        )

        self.assertEqual(fields, {'acl': 'public-read'})

        _, post_kwargs = self.presign_post_mock.call_args
        request_dict = post_kwargs['request_dict']
        fields = post_kwargs['fields']
        conditions = post_kwargs['conditions']
        self.assertEqual(
            request_dict['url'], 'https://mybucket.s3.amazonaws.com/'
        )
        self.assertEqual(
            conditions,
            [{'acl': 'public-read'}, {'bucket': 'mybucket'}, {'key': 'mykey'}],
        )
        self.assertEqual(fields['acl'], 'public-read')
        self.assertEqual(fields, {'key': 'mykey', 'acl': 'public-read'})

    def test_generate_presigned_post_non_s3_client(self):
        self.client = self.session.create_client('ec2', 'us-west-2')
        with self.assertRaises(AttributeError):
            self.client.generate_presigned_post()


class TestGenerateDBAuthToken(BaseSignerTest):
    maxDiff = None

    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client(
            'rds',
            region_name='us-east-1',
            aws_access_key_id='akid',
            aws_secret_access_key='skid',
            config=Config(signature_version='v4'),
        )

    def test_generate_db_auth_token(self):
        hostname = 'prod-instance.us-east-1.rds.amazonaws.com'
        port = 3306
        username = 'someusername'
        clock = datetime.datetime(2016, 11, 7, 17, 39, 33, tzinfo=tzutc())

        with mock.patch('datetime.datetime') as dt:
            dt.now.return_value = clock
            result = generate_db_auth_token(
                self.client, hostname, port, username
            )

        expected_result = (
            'prod-instance.us-east-1.rds.amazonaws.com:3306/?Action=connect'
            '&DBUser=someusername&X-Amz-Algorithm=AWS4-HMAC-SHA256'
            '&X-Amz-Date=20161107T173933Z&X-Amz-SignedHeaders=host'
            '&X-Amz-Expires=900&X-Amz-Credential=akid%2F20161107%2F'
            'us-east-1%2Frds-db%2Faws4_request&X-Amz-Signature'
            '=d1138cdbc0ca63eec012ec0fc6c2267e03642168f5884a7795320d4c18374c61'
        )

        # A scheme needs to be appended to the beginning or urlsplit may fail
        # on certain systems.
        assert_url_equal('https://' + result, 'https://' + expected_result)

    def test_custom_region(self):
        hostname = 'host.us-east-1.rds.amazonaws.com'
        port = 3306
        username = 'mySQLUser'
        region = 'us-west-2'
        result = generate_db_auth_token(
            self.client, hostname, port, username, Region=region
        )

        self.assertIn(region, result)
        # The hostname won't be changed even if a different region is specified
        self.assertIn(hostname, result)


class TestDSQLGenerateDBAuthToken(BaseSignerTest):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client(
            'dsql',
            region_name='us-east-1',
            aws_access_key_id='ACCESS_KEY',
            aws_secret_access_key='SECRET_KEY',
            aws_session_token="SESSION_TOKEN",
        )
        self.hostname = 'test.dsql.us-east-1.on.aws'
        self.action = 'DbConnect'

    @FreezeTime(botocore.auth.datetime, date=DATE)
    def test_dsql_generate_db_auth_token(self):
        result = _dsql_generate_db_auth_token(
            self.client, self.hostname, self.action
        )

        expected_result = (
            'test.dsql.us-east-1.on.aws/?Action=DbConnect'
            '&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential='
            'ACCESS_KEY%2F20241107%2Fus-east-1%2Fdsql%2Faws4_request'
            '&X-Amz-Date=20241107T173933Z&X-Amz-Expires=900&X-Amz-SignedHeaders=host'
            '&X-Amz-Security-Token=SESSION_TOKEN&X-Amz-Signature='
            '57fe03e060348aaa21405c239bf02572bbc911076e94dcd65c12ae569dd8fcf4'
        )

        # A scheme needs to be appended to the beginning or urlsplit may fail
        # on certain systems.
        assert_url_equal('https://' + result, 'https://' + expected_result)

    @FreezeTime(botocore.auth.datetime, date=DATE)
    def test_dsql_generate_db_connect_auth_token(self):
        result = dsql_generate_db_connect_auth_token(
            self.client, self.hostname
        )

        expected_result = (
            'test.dsql.us-east-1.on.aws/?Action=DbConnect'
            '&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential='
            'ACCESS_KEY%2F20241107%2Fus-east-1%2Fdsql%2Faws4_request'
            '&X-Amz-Date=20241107T173933Z&X-Amz-Expires=900&X-Amz-SignedHeaders=host'
            '&X-Amz-Security-Token=SESSION_TOKEN&X-Amz-Signature='
            '57fe03e060348aaa21405c239bf02572bbc911076e94dcd65c12ae569dd8fcf4'
        )

        # A scheme needs to be appended to the beginning or urlsplit may fail
        # on certain systems.
        assert_url_equal('https://' + result, 'https://' + expected_result)

    @FreezeTime(botocore.auth.datetime, date=DATE)
    def test_dsql_generate_db_connect_admin_auth_token(self):
        result = dsql_generate_db_connect_admin_auth_token(
            self.client, self.hostname
        )

        expected_result = (
            'test.dsql.us-east-1.on.aws/?Action=DbConnectAdmin'
            '&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential='
            'ACCESS_KEY%2F20241107%2Fus-east-1%2Fdsql%2Faws4_request'
            '&X-Amz-Date=20241107T173933Z&X-Amz-Expires=900&X-Amz-SignedHeaders=host'
            '&X-Amz-Security-Token=SESSION_TOKEN&X-Amz-Signature='
            '5ac084bc7cabccc19a52a5d1b5c24b50d3ce143f43b659bd484c91aaf555e190'
        )

        # A scheme needs to be appended to the beginning or urlsplit may fail
        # on certain systems.
        assert_url_equal('https://' + result, 'https://' + expected_result)

    def test_dsql_generate_db_auth_token_invalid_action(self):
        with self.assertRaises(ParamValidationError):
            _dsql_generate_db_auth_token(self.client, self.hostname, "FooBar")


@pytest.mark.parametrize(
    'request_method',
    ['GET', 'HEAD', 'OPTIONS', 'POST', 'PUT', 'DELETE', None],
)
def test_generate_presigned_url_content_type_removal_for_polly(
    polly_client,
    request_method,
):
    url = polly_client.generate_presigned_url(
        'synthesize_speech',
        Params={
            'OutputFormat': 'mp3',
            'Text': 'Hello world!',
            'VoiceId': 'Joanna',
        },
        HttpMethod=request_method,
    )
    assert 'content-type' not in url.lower()



"""Smoke tests to verify basic communication to all AWS services.

If you want to control what services/regions are used you can
also provide two separate env vars:

    * AWS_SMOKE_TEST_REGION - The region used to create clients.
    * AWS_SMOKE_TEST_SERVICES - A CSV list of service names to test.

Otherwise, the ``REGION`` variable specifies the default region
to use and all the services in SMOKE_TESTS/ERROR_TESTS will be tested.

"""

import logging
import os
import warnings
from pprint import pformat

import pytest

import botocore.session
from botocore import xform_name
from botocore.client import ClientError
from botocore.exceptions import ConnectionClosedError
from tests import ClientHTTPStubber

# Mapping of service -> api calls to try.
# Each api call is a dict of OperationName->params.
# Empty params means that the operation will be called with no params.  This is
# used as a quick verification that we can successfully make calls to services.
SMOKE_TESTS = {
    'acm': {'ListCertificates': {}},
    'apigateway': {'GetRestApis': {}},
    'autoscaling': {
        'DescribeAccountLimits': {},
        'DescribeAdjustmentTypes': {},
    },
    'cloudformation': {'DescribeStacks': {}, 'ListStacks': {}},
    'cloudfront': {'ListDistributions': {}, 'ListStreamingDistributions': {}},
    'cloudhsmv2': {'DescribeBackups': {}},
    'cloudtrail': {'DescribeTrails': {}},
    'cloudwatch': {'ListMetrics': {}},
    'codecommit': {'ListRepositories': {}},
    'codedeploy': {'ListApplications': {}},
    'codepipeline': {'ListActionTypes': {}},
    'cognito-identity': {'ListIdentityPools': {'MaxResults': 1}},
    'cognito-sync': {'ListIdentityPoolUsage': {}},
    'config': {'DescribeDeliveryChannels': {}},
    'datapipeline': {'ListPipelines': {}},
    'devicefarm': {'ListProjects': {}},
    'directconnect': {'DescribeConnections': {}},
    'ds': {'DescribeDirectories': {}},
    'dynamodb': {'ListTables': {}},
    'dynamodbstreams': {'ListStreams': {}},
    'ec2': {'DescribeRegions': {}, 'DescribeInstances': {}},
    'ecr': {'DescribeRepositories': {}},
    'ecs': {'DescribeClusters': {}},
    'elasticache': {'DescribeCacheClusters': {}},
    # 'elasticbeanstalk': {'DescribeApplications': {}},
    'elb': {'DescribeLoadBalancers': {}},
    'emr': {'ListClusters': {}},
    'es': {'ListDomainNames': {}},
    'events': {'ListRules': {}},
    'firehose': {'ListDeliveryStreams': {}},
    'gamelift': {'ListBuilds': {}},
    'iam': {'ListUsers': {}},
    # Does not work with session credentials so
    # importexport tests are not run.
    # 'importexport': {'ListJobs': {}},
    'importexport': {},
    'iot': {'DescribeEndpoint': {}},
    'kinesis': {'ListStreams': {}},
    'kms': {'ListKeys': {}},
    'lambda': {'ListFunctions': {}},
    'logs': {'DescribeLogGroups': {}},
    'rds': {'DescribeDBInstances': {}},
    'redshift': {'DescribeClusters': {}},
    # 'route53': {'ListHostedZones': {}},
    'route53domains': {'ListDomains': {}},
    's3': {'ListBuckets': {}},
    'sdb': {'ListDomains': {}},
    'ses': {'ListIdentities': {}},
    'shield': {'GetSubscriptionState': {}},
    'sns': {'ListTopics': {}},
    'sqs': {'ListQueues': {}},
    'ssm': {'ListDocuments': {}},
    # sts tests would normally go here, but
    # there aren't any calls you can make when
    # using session credentials so we don't run any
    # sts tests.
    'sts': {},
    # 'sts': {'GetSessionToken': {}},
    # Subscription needed for support API calls.
    'support': {},
    'swf': {'ListDomains': {'registrationStatus': 'REGISTERED'}},
    'waf': {'ListWebACLs': {'Limit': 1}},
    'workspaces': {'DescribeWorkspaces': {}},
}


# Same thing as the SMOKE_TESTS hash above, except these verify
# that we get an error response back from the server because
# we've sent invalid params.
ERROR_TESTS = {
    'apigateway': {'GetRestApi': {'restApiId': 'fake-id'}},
    'autoscaling': {
        'CreateLaunchConfiguration': {
            'LaunchConfigurationName': 'foo',
            'ImageId': 'ami-12345678',
            'InstanceType': 'm1.small',
        }
    },
    'cloudformation': {
        'CreateStack': {
            'StackName': 'fake',
            'TemplateURL': 'http://s3.amazonaws.com/foo/bar',
        }
    },
    'cloudfront': {'GetDistribution': {'Id': 'fake-id'}},
    'cloudhsmv2': {'ListTags': {'ResourceId': 'fake-id'}},
    'cloudtrail': {'DeleteTrail': {'Name': 'fake-trail'}},
    'cloudwatch': {
        'SetAlarmState': {
            'AlarmName': 'abc',
            'StateValue': 'mno',
            'StateReason': 'xyz',
        }
    },
    'logs': {'GetLogEvents': {'logGroupName': 'a', 'logStreamName': 'b'}},
    'codecommit': {'ListBranches': {'repositoryName': 'fake-repo'}},
    'codedeploy': {'GetDeployment': {'deploymentId': 'fake-id'}},
    'codepipeline': {'GetPipeline': {'name': 'fake-pipeline'}},
    'cognito-identity': {'DescribeIdentityPool': {'IdentityPoolId': 'fake'}},
    'cognito-sync': {'DescribeIdentityPoolUsage': {'IdentityPoolId': 'fake'}},
    'config': {
        'GetResourceConfigHistory': {'resourceType': '', 'resourceId': 'fake'},
    },
    'datapipeline': {'GetPipelineDefinition': {'pipelineId': 'fake'}},
    'devicefarm': {
        'GetDevice': {'arn': 'arn:aws:devicefarm:REGION::device:f'}
    },
    'directconnect': {'DescribeConnections': {'connectionId': 'fake'}},
    'ds': {'CreateDirectory': {'Name': 'n', 'Password': 'p', 'Size': '1'}},
    'dynamodb': {'DescribeTable': {'TableName': 'fake'}},
    'dynamodbstreams': {'DescribeStream': {'StreamArn': 'x' * 37}},
    'ec2': {'DescribeInstances': {'InstanceIds': ['i-12345678']}},
    'ecs': {'StopTask': {'task': 'fake'}},
    'efs': {'DeleteFileSystem': {'FileSystemId': 'fake'}},
    'elasticache': {'DescribeCacheClusters': {'CacheClusterId': 'fake'}},
    # 'elasticbeanstalk': {
    #     'DescribeEnvironmentResources': {'EnvironmentId': 'x'},
    # },
    'elb': {'DescribeLoadBalancers': {'LoadBalancerNames': ['fake']}},
    'emr': {'DescribeCluster': {'ClusterId': 'fake'}},
    'es': {'DescribeElasticsearchDomain': {'DomainName': 'not-a-domain'}},
    'gamelift': {'DescribeBuild': {'BuildId': 'fake-build-id'}},
    'iam': {'GetUser': {'UserName': 'fake'}},
    'kinesis': {'DescribeStream': {'StreamName': 'fake'}},
    'kms': {'GetKeyPolicy': {'KeyId': 'fake', 'PolicyName': 'fake'}},
    'lambda': {'Invoke': {'FunctionName': 'fake'}},
    'rds': {'DescribeDBInstances': {'DBInstanceIdentifier': 'fake'}},
    'redshift': {'DescribeClusters': {'ClusterIdentifier': 'fake'}},
    # 'route53': {'GetHostedZone': {'Id': 'fake'}},
    'route53domains': {'GetDomainDetail': {'DomainName': 'fake'}},
    's3': {'ListObjects': {'Bucket': 'thisbucketdoesnotexistasdf'}},
    'ses': {'VerifyEmailIdentity': {'EmailAddress': 'fake'}},
    'sdb': {'CreateDomain': {'DomainName': ''}},
    'sns': {
        'ConfirmSubscription': {'TopicArn': 'a', 'Token': 'b'},
        'Publish': {'Message': 'hello', 'TopicArn': 'fake'},
    },
    'sqs': {'GetQueueUrl': {'QueueName': 'fake'}},
    'ssm': {'GetDocument': {'Name': 'fake'}},
    'storagegateway': {'ListVolumes': {'GatewayARN': 'x' * 50}},
    'sts': {'GetFederationToken': {'Name': 'fake', 'Policy': 'fake'}},
    'support': {
        'CreateCase': {
            'subject': 'x',
            'communicationBody': 'x',
            'categoryCode': 'x',
            'serviceCode': 'x',
            'severityCode': 'low',
        }
    },
    'swf': {'DescribeDomain': {'name': 'fake'}},
    'waf': {'GetWebACL': {'WebACLId': 'fake'}},
    'workspaces': {'DescribeWorkspaces': {'DirectoryId': 'fake-directory-id'}},
}

REGION = 'us-east-1'
REGION_OVERRIDES = {
    'devicefarm': 'us-west-2',
    'efs': 'us-west-2',
}
MAX_RETRIES = 8
logger = logging.getLogger(__name__)


def _get_client(session, service):
    if os.environ.get('AWS_SMOKE_TEST_REGION', ''):
        region_name = os.environ['AWS_SMOKE_TEST_REGION']
    else:
        region_name = REGION_OVERRIDES.get(service, REGION)
    client = session.create_client(service, region_name=region_name)
    client.meta.events.register_first('needs-retry.*.*', retry_handler)
    return client


def retry_handler(response, attempts, **kwargs):
    if response is not None:
        _, parsed = response
        code = parsed.get('Error', {}).get('Code')
        # Catch ThrottleException, Throttling.
        is_throttle_error = code is not None and 'throttl' in code.lower()
        if is_throttle_error and attempts <= MAX_RETRIES:
            # We want the exponential behavior with a fixed 10 second
            # minimum, e.g. 11, 12, 14, 18, 26.  With a max retries of 8,
            # this is about 7-8 minutes total we'll retry.
            retry_delay = (2 ** (attempts - 1)) + 10
            logger.debug("Using custom retry delay of: %s", retry_delay)
            return retry_delay


def _list_services(dict_entries):
    # List all services in the provided dict_entry.
    # If the AWS_SMOKE_TEST_SERVICES is provided,
    # it's a comma separated list of services you can provide
    # if you only want to run the smoke tests for certain services.
    if 'AWS_SMOKE_TEST_SERVICES' not in os.environ:
        return dict_entries.keys()
    else:
        wanted_services = os.environ.get('AWS_SMOKE_TEST_SERVICES', '').split(
            ','
        )
        return [key for key in dict_entries if key in wanted_services]


@pytest.fixture()
def botocore_session():
    return botocore.session.get_session()


def _smoke_tests():
    for service_name in _list_services(SMOKE_TESTS):
        for operation_name in SMOKE_TESTS[service_name]:
            kwargs = SMOKE_TESTS[service_name][operation_name]
            yield service_name, operation_name, kwargs


def _error_tests():
    for service_name in _list_services(ERROR_TESTS):
        for operation_name in ERROR_TESTS[service_name]:
            kwargs = ERROR_TESTS[service_name][operation_name]
            yield service_name, operation_name, kwargs


@pytest.mark.parametrize(
    "service_name, operation_name, kwargs", _smoke_tests()
)
def test_can_make_request_with_client(
    botocore_session, service_name, operation_name, kwargs
):
    # Same as test_can_make_request, but with Client objects
    # instead of service/operations.
    client = _get_client(botocore_session, service_name)
    method = getattr(client, xform_name(operation_name))
    with warnings.catch_warnings(record=True) as caught_warnings:
        response = method(**kwargs)
        err_msg = f"Warnings were emitted during smoke test: {caught_warnings}"
        assert len(caught_warnings) == 0, err_msg
        assert 'Errors' not in response


@pytest.mark.parametrize(
    "service_name, operation_name, kwargs", _error_tests()
)
def test_can_make_request_and_understand_errors_with_client(
    botocore_session, service_name, operation_name, kwargs
):
    client = _get_client(botocore_session, service_name)
    method = getattr(client, xform_name(operation_name))
    with pytest.raises(ClientError):
        method(**kwargs)


@pytest.mark.parametrize(
    "service_name, operation_name, kwargs", _smoke_tests()
)
def test_client_can_retry_request_properly(
    botocore_session, service_name, operation_name, kwargs
):
    client = _get_client(botocore_session, service_name)
    operation = getattr(client, xform_name(operation_name))
    exception = ConnectionClosedError(endpoint_url='https://mock.eror')
    with ClientHTTPStubber(client, strict=False) as http_stubber:
        http_stubber.responses.append(exception)
        try:
            operation(**kwargs)
        except ClientError as e:
            assert False, (
                'Request was not retried properly, '
                f'received error:\n{pformat(e)}'
            )
        # Ensure we used the stubber as we're not using it in strict mode
        assert len(http_stubber.responses) == 0, 'Stubber was not used!'
